"""A bunch of fitted opsin models."""

from brian2.units import (
    mM,
    mm2,
    ms,
    mV,
    nsiemens,
    pcoulomb,
    psiemens,
    second,
)

from cleo.light.light_dependence import equal_photon_flux_spectrum, plot_spectra
from cleo.opto.opsins import (
    BansalFourStateOpsin,
    BansalThreeStatePump,
    FourStateOpsin,
)


ONE_P_TWO_P_RATIO = 270382.2996938772


def chr2_4s() -> FourStateOpsin:
    """Returns a 4-state ChR2 model.

    Params taken from try.projectpyrho.org's default 4-state configuration.
    Action spectrum from `Nagel et al., 2003, Fig. 4a
    <https://www.pnas.org/doi/full/10.1073/pnas.1936192100>`_,
    extracted using `Plot Digitizer <https://plotdigitizer.com/>`_.

    Parameters can be changed after initialization but *before injection*.
    """
    return FourStateOpsin(
        g0=114000 * psiemens,
        gamma=0.00742,
        phim=2.33e17 / mm2 / second,  # *photon, not in Brian2
        k1=4.15 / ms,
        k2=0.868 / ms,
        p=0.833,
        Gf0=0.0373 / ms,
        kf=0.0581 / ms,
        Gb0=0.0161 / ms,
        kb=0.063 / ms,
        q=1.94,
        Gd1=0.105 / ms,
        Gd2=0.0138 / ms,
        Gr0=0.00033 / ms,
        E=0 * mV,
        v0=43 * mV,
        v1=17.1 * mV,
        name="ChR2",
        spectrum=[
            (400, 0.34),
            (422, 0.65),
            (460, 0.96),
            (470, 1),
            (473, 1),
            (500, 0.57),
            (520, 0.22),
            (540, 0.06),
            (560, 0.01),
            (800, 0.34 / ONE_P_TWO_P_RATIO),
            (844, 0.65 / ONE_P_TWO_P_RATIO),
            (920, 0.96 / ONE_P_TWO_P_RATIO),
            (940, 1 / ONE_P_TWO_P_RATIO),
            (946, 1 / ONE_P_TWO_P_RATIO),
            (1000, 0.57 / ONE_P_TWO_P_RATIO),
            (1040, 0.22 / ONE_P_TWO_P_RATIO),
            (1080, 0.06 / ONE_P_TWO_P_RATIO),
            (1120, 0.01 / ONE_P_TWO_P_RATIO),
        ],
    )


def chr2_b4s() -> BansalFourStateOpsin:
    """Returns a 4-state ChR2 model.

    Params given in Bansal et al., 2020.
    Action spectrum from `Nagel et al., 2003, Fig. 4a
    <https://www.pnas.org/doi/full/10.1073/pnas.1936192100>`_,
    extracted using `Plot Digitizer <https://plotdigitizer.com/>`_.

    Parameters can be changed after initialization but *before injection*.
    """
    return BansalFourStateOpsin(
        Gd1=0.066 / ms,
        Gd2=0.01 / ms,
        Gr0=3.33e-4 / ms,
        g0=3.2 * nsiemens,
        phim=1e16 / mm2 / second,  # *photon, not in Brian2
        k1=0.4 / ms,
        k2=0.12 / ms,
        Gf0=0.018 / ms,
        Gb0=0.008 / ms,
        kf=0.01 / ms,
        kb=0.008 / ms,
        gamma=0.05,
        p=1,
        q=1,
        E=0 * mV,
        name="ChR2",
        spectrum=[
            (400, 0.34),
            (422, 0.65),
            (460, 0.96),
            (470, 1),
            (473, 1),
            (500, 0.57),
            (520, 0.22),
            (540, 0.06),
            (560, 0.01),
            (800, 0.34 / ONE_P_TWO_P_RATIO),
            (844, 0.65 / ONE_P_TWO_P_RATIO),
            (920, 0.96 / ONE_P_TWO_P_RATIO),
            (940, 1 / ONE_P_TWO_P_RATIO),
            (946, 1 / ONE_P_TWO_P_RATIO),
            (1000, 0.57 / ONE_P_TWO_P_RATIO),
            (1040, 0.22 / ONE_P_TWO_P_RATIO),
            (1080, 0.06 / ONE_P_TWO_P_RATIO),
            (1120, 0.01 / ONE_P_TWO_P_RATIO),
        ],
    )


def chr2_h134r_4s() -> BansalFourStateOpsin:
    """Returns a 4-state ChR2(H134R) model.

    Params given in Bansal et al., 2020.
    Action spectrum is same as for :func:`~chr2_4s`, but blue-shifted 20 nm
    (I cannot find it directly in the literature).

    Parameters can be changed after initialization but *before injection*.
    """

    return BansalFourStateOpsin(
        Gd1=0.045 / ms,
        Gd2=0.01 / ms,
        Gr0=1e-4 / ms,
        g0=37.2 * nsiemens,
        phim=1e16 / mm2 / second,  # *photon, not in Brian2
        k1=0.3 / ms,
        k2=0.12 / ms,
        Gf0=0.014 / ms,
        Gb0=0.01 / ms,
        kf=0.01 / ms,
        kb=0.01 / ms,
        gamma=0.05,
        p=1,
        q=1,
        E=0 * mV,
        name="ChR2(H134R)",
        spectrum=[
            (380, 0.34),
            (402, 0.65),
            (440, 0.96),
            (450, 1),
            (480, 0.57),
            (500, 0.22),
            (520, 0.06),
            (540, 0.01),
            # TODO: 40 nm shift?
            (760, 0.34 / ONE_P_TWO_P_RATIO),
            (804, 0.65 / ONE_P_TWO_P_RATIO),
            (880, 0.96 / ONE_P_TWO_P_RATIO),
            (900, 1 / ONE_P_TWO_P_RATIO),
            (960, 0.57 / ONE_P_TWO_P_RATIO),
            (1000, 0.22 / ONE_P_TWO_P_RATIO),
            (1040, 0.06 / ONE_P_TWO_P_RATIO),
            (1080, 0.01 / ONE_P_TWO_P_RATIO),
        ],
    )


def vfchrimson_4s() -> BansalFourStateOpsin:
    """Returns a 4-state vf-Chrimson model.

    Params given in Bansal et al., 2020.
    Action spectrum from `Mager et al., 2018, Supp. Fig. 1a
    <https://static-content.springer.com/esm/art%3A10.1038%2Fs41467-018-04146-3/MediaObjects/41467_2018_4146_MOESM1_ESM.docx>`_,
    extracted using `Plot Digitizer <https://plotdigitizer.com/>`_.

    Parameters can be changed after initialization but *before injection*.
    """
    return BansalFourStateOpsin(
        Gd1=0.37 / ms,
        Gd2=0.175 / ms,
        Gr0=6.67e-7 / ms,
        g0=17.5 * nsiemens,
        phim=1.5e16 / mm2 / second,
        k1=3 / ms,
        k2=0.2 / ms,
        Gf0=0.02 / ms,
        Gb0=3.2e-3 / ms,
        kf=0.01 / ms,
        kb=0.01 / ms,
        gamma=0.05,
        p=1,
        q=1,
        E=0 * mV,
        name="vf-Chrimson",
        spectrum=equal_photon_flux_spectrum(
            [
                (470, 0.34),
                (490, 0.51),
                (510, 0.71),
                (530, 0.75),
                (550, 0.86),
                (570, 1),
                (590, 1),
                (610, 0.8),
                (630, 0.48),
                (940, 0.34 / ONE_P_TWO_P_RATIO),
                (980, 0.51 / ONE_P_TWO_P_RATIO),
                (1020, 0.71 / ONE_P_TWO_P_RATIO),
                (1060, 0.75 / ONE_P_TWO_P_RATIO),
                (1100, 0.86 / ONE_P_TWO_P_RATIO),
                (1140, 1 / ONE_P_TWO_P_RATIO),
                (1180, 1 / ONE_P_TWO_P_RATIO),
                (1220, 0.8 / ONE_P_TWO_P_RATIO),
                (1260, 0.48 / ONE_P_TWO_P_RATIO),
            ]
        ),
    )


def chrimson_4s() -> BansalFourStateOpsin:
    """Returns a 4-state Chrimson model.

    Params given in Bansal et al., 2020.
    Action spectrum from `Mager et al., 2018, Supp. Fig. 1a
    <https://static-content.springer.com/esm/art%3A10.1038%2Fs41467-018-04146-3/MediaObjects/41467_2018_4146_MOESM1_ESM.docx>`_,
    extracted using `Plot Digitizer <https://plotdigitizer.com/>`_.

    Parameters can be changed after initialization but *before injection*.
    """
    return BansalFourStateOpsin(
        Gd1=0.041 / ms,
        Gd2=0.01 / ms,
        Gr0=6.67e-7 / ms,
        g0=22 * nsiemens,
        phim=0.2e15 / mm2 / second,
        k1=0.05 / ms,
        k2=0.5 / ms,
        Gf0=0.001 / ms,
        Gb0=0.01 / ms,
        kf=0.01 / ms,
        kb=0.004 / ms,
        gamma=0.05,
        p=1,
        q=1,
        E=0 * mV,
        name="Chrimson",
        spectrum=equal_photon_flux_spectrum(
            [
                (470, 0.31),
                (490, 0.47),
                (510, 0.69),
                (530, 0.75),
                (550, 0.88),
                (570, 0.97),
                (590, 1),
                (610, 0.88),
                (630, 0.55),
                (940, 0.31 / ONE_P_TWO_P_RATIO),
                (980, 0.47 / ONE_P_TWO_P_RATIO),
                (1020, 0.69 / ONE_P_TWO_P_RATIO),
                (1060, 0.75 / ONE_P_TWO_P_RATIO),
                (1100, 0.88 / ONE_P_TWO_P_RATIO),
                (1140, 0.97 / ONE_P_TWO_P_RATIO),
                (1180, 1 / ONE_P_TWO_P_RATIO),
                (1220, 0.88 / ONE_P_TWO_P_RATIO),
                (1260, 0.55 / ONE_P_TWO_P_RATIO),
            ]
        ),
    )


def gtacr2_4s() -> BansalFourStateOpsin:
    """Returns a 4-state model of GtACR2, an anion channel.

    Params given in Bansal et al., 2020.
    Action spectra from `Govorunova et al., 2015, Fig. 1f
    <https://www.science.org/doi/10.1126/science.aaa7484#F1>`_,
    extracted using `Plot Digitizer <https://plotdigitizer.com/>`_.

    Parameters can be changed after initialization but *before injection*.
    """
    return BansalFourStateOpsin(
        Gd1=0.017 / ms,
        Gd2=0.01 / ms,
        Gr0=5.8e-4 / ms,
        g0=44 * nsiemens,
        phim=2e17 / mm2 / second,
        k1=40 / ms,
        k2=20 / ms,
        Gf0=0.001 / ms,
        Gb0=0.003 / ms,
        kf=0.001 / ms,
        kb=0.005 / ms,
        gamma=0.05,
        p=1,
        q=0.1,
        E=-69.5 * mV,
        name="GtACR2",
        spectrum=[
            (400, 0.40),
            (410, 0.49),
            (420, 0.56),
            (430, 0.65),
            (440, 0.82),
            (450, 0.88),
            (460, 0.88),
            (470, 1.0),
            (480, 0.91),
            (490, 0.67),
            (500, 0.41),
            (510, 0.21),
            (520, 0.12),
            (530, 0.06),
            (540, 0.02),
            (550, 0.00),
            (560, 0.00),
            (800, 0.4 / ONE_P_TWO_P_RATIO),
            (820, 0.49 / ONE_P_TWO_P_RATIO),
            (840, 0.56 / ONE_P_TWO_P_RATIO),
            (860, 0.65 / ONE_P_TWO_P_RATIO),
            (880, 0.82 / ONE_P_TWO_P_RATIO),
            (900, 0.88 / ONE_P_TWO_P_RATIO),
            (920, 0.88 / ONE_P_TWO_P_RATIO),
            (940, 1.0 / ONE_P_TWO_P_RATIO),
            (960, 0.91 / ONE_P_TWO_P_RATIO),
            (980, 0.67 / ONE_P_TWO_P_RATIO),
            (1000, 0.41 / ONE_P_TWO_P_RATIO),
            (1020, 0.21 / ONE_P_TWO_P_RATIO),
            (1040, 0.12 / ONE_P_TWO_P_RATIO),
            (1060, 0.06 / ONE_P_TWO_P_RATIO),
            (1080, 0.02 / ONE_P_TWO_P_RATIO),
            (1100, 0.0 / ONE_P_TWO_P_RATIO),
            (1120, 0.0 / ONE_P_TWO_P_RATIO),
        ],
    )


def enphr3_3s():
    """Returns a 3-state model of eNpHR3, a chloride pump.

    Params given in Bansal et al., 2020.
    Action spectrum from `Gradinaru et al., 2010 <https://doi.org/10.1016/j.cell.2010.02.037>`_,
    Figure 3F,
    extracted using `Plot Digitizer <https://plotdigitizer.com/>`_.
    """
    return BansalThreeStatePump(
        Gd=0.025 / ms,
        Gr=0.05 / ms,
        ka=1 / ms,
        p=0.7,
        q=0.1,
        phim=1.2e18 / mm2 / second,
        E=-400 * mV,
        g0=22.34 * nsiemens,
        a=0.02e-2 * mM / pcoulomb,
        b=12,
        name="eNpHR3.0",
        spectrum=[
            (390, 0.162),
            (405, 0.239),
            (430, 0.255),
            (445, 0.255),
            (470, 0.371),
            (495, 0.554),
            (520, 0.716),
            (542.5, 0.840),
            (560, 0.930),
            (590, 1),
            (630, 0.385),
            (780, 0.162 / ONE_P_TWO_P_RATIO),
            (810, 0.239 / ONE_P_TWO_P_RATIO),
            (860, 0.255 / ONE_P_TWO_P_RATIO),
            (890, 0.255 / ONE_P_TWO_P_RATIO),
            (940, 0.371 / ONE_P_TWO_P_RATIO),
            (990, 0.554 / ONE_P_TWO_P_RATIO),
            (1040, 0.716 / ONE_P_TWO_P_RATIO),
            (1085.0, 0.84 / ONE_P_TWO_P_RATIO),
            (1120, 0.93 / ONE_P_TWO_P_RATIO),
            (1180, 1 / ONE_P_TWO_P_RATIO),
            (1260, 0.385 / ONE_P_TWO_P_RATIO),
        ],
    )


if __name__ == "__main__":
    import matplotlib.pyplot as plt

    plot_spectra(
        chr2_4s(),
        chr2_h134r_4s(),
        vfchrimson_4s(),
        chrimson_4s(),
        gtacr2_4s(),
        enphr3_3s(),
        extrapolate=True,
        range="1p2p",
    )
    plt.show()



"""Contains multi-unit and sorted spiking signals."""

from __future__ import annotations

import warnings
from abc import abstractmethod
from datetime import datetime
from functools import cache
from typing import Callable, Tuple

import brian2.only as b2
import neo
import numpy as np
import quantities as pq
from attrs import define, field, fields
from brian2 import NeuronGroup, Quantity, SpikeMonitor, mm, ms, um
from jaxtyping import Bool, Float, Int
from scipy import signal
from scipy.stats import norm

from cleo.base import NeoExportable
from cleo.ephys.probes import Signal
from cleo.utilities import rng, unit_safe_cat


@define(eq=False)
class Spiking(Signal, NeoExportable):
    """Base class for probabilistically detecting spikes.

    See ``notebooks/spike_detection.py`` for an interactive explanation of the methods
    and parameters involved."""

    r_noise_floor: Quantity = 80 * um
    """Radius (with Brian unit) at which the measured spike amplitude
    equals the background noise standard deviation. 
    i.e., ``spike_amplitude(r_noise_floor) = sigma_noise = 1``.
    80 μm by default."""
    threshold_sigma: int = 4
    """Spike detection threshold, as a multiple of sigma_noise. 
    Values in real experiments typically range from 3 to 6. 4 by default."""
    spike_amplitude_cv: float = 0.05
    """Coefficient of variation of the spike amplitude, i.e., `|sigma_amp/mu_amp|`. 
    From what we have seen in Allen Cell Types data, this ranges from 0 to 0.2, 
    but is most often very low. 0.05 by default."""
    r0: Quantity = 5 * um
    """A small distance added to r before computing the amplitude to avoid division
    by 0 for the power law decay. 5 μm by default.
    
    It also makes some physical sense as the minimum distance from the current source
    it is possible to place an electrode, 5 μm being reasonable as the radius of a typical soma."""
    recording_recall_cutoff: float = 0.001
    """*Multi-channel* recall, above which neurons will be considered.
    I.e., the probability a spike is detected on at least one channel.
    
    You shouldn't need to change this; it's mainly for efficiency, allowing amplitude 
    sampling and threshold crossing to operate on fewer spikes by ignoring neurons
    very unlikely to produce a spike that crosses the threshold."""
    eap_decay_fn: Callable[[Quantity], float] = lambda r: r**-2
    """The function describing the decay of the measured extracellular action potential
    amplitude. By default 1/r^2.
    
    This inverse square decay is a good approximation in accordance with the detailed
    simulations by `Pettersen et al. (2008) <https://www.ncbi.nlm.nih.gov/pmc/articles/PMC2186261/>`_,
    though they find the exponent ranges from 2 to 3 depending on the cell type and distance. """
    collision_prob_fn: Callable[[Quantity], float] = field(default=None)
    """The probability of failing to detect the latter of two overlapping threshold
    crossings on a given channel, as a function of ``t2 - t1``.
    Values for ``t2 - t1 < 0`` are ignored.

    For :class:`SortedSpiking`, the default is a decaying exponential.
    See `Garcia et al. (2022) <https://www.eneuro.org/content/9/5/ENEURO.0105-22.2022>`_
    for what this might look like for different sorters.

    By default simply enforces a hard 1 ms refractory period
    per channel for :class:`MultiUnitActivity`.
    """

    @collision_prob_fn.validator
    def _validate_coll_prob_fn(self, attribute, value):
        if value is not None:
            assert callable(value), "collision_prob_fn must be callable"
            assert np.all(0 <= value([0, 1, 10] * ms) <= 1), (
                "collision_prob_fn must return a value between 0 and 1"
            )

    simulate_false_positives: bool = True
    """Whether to simulate false positives from noise. In the case of :class:`SortedSpiking`,
    these aren't reported, but still affect collision sampling."""

    t: Quantity = field(
        init=False, factory=lambda: ms * np.array([], dtype=float), repr=False
    )
    """Spike times with Brian units, stored if
    :attr:`~cleo.InterfaceDevice.save_history` on :attr:`~Signal.probe`"""
    i: Int[np.ndarray, "n_recorded_spikes"] = field(
        init=False, factory=lambda: np.array([], dtype=int), repr=False
    )
    """Channel (for multi-unit) or neuron (for sorted) indices
    of spikes, stored if
    :attr:`~cleo.InterfaceDevice.save_history` on :attr:`~Signal.probe`"""
    t_samp: Quantity = field(
        init=False, factory=lambda: ms * np.array([], dtype=float), repr=False
    )
    """Sample times with Brian units when each spike was recorded, stored if
    :attr:`~cleo.InterfaceDevice.save_history` on :attr:`~Signal.probe`"""
    i_probe_by_ng: dict[NeuronGroup, Int[np.ndarray, "ng_N"]] = field(
        init=False, factory=dict, repr=False
    )
    """neuron_group keys, i_probe values for every neuron in group."""
    i_ng_by_i_probe: list[tuple[NeuronGroup, int]] = field(
        init=False, factory=list, repr=False
    )
    """n_neurons-length list indexed by i_probe returning a neuron_group, i_ng tuple
    to map from i_probe to neuron group and index."""
    _monitors: list[SpikeMonitor] = field(init=False, factory=list, repr=False)
    _mon_spikes_already_seen: list[int] = field(init=False, factory=list, repr=False)
    _mu_eap: Float[np.ndarray, "n_neurons n_channels"] = field(
        init=False, default=None, repr=False
    )
    _prev_t: Quantity = field(init=False, default=None, repr=False)
    _prev_zi: Float[np.ndarray, "n_sos 2"] = field(init=False, default=None, repr=False)

    def _init_saved_vars(self):
        if self.probe.save_history:
            self.t = fields(type(self)).t.default.factory()
            self.i = fields(type(self)).i.default.factory()
            self.t_samp = fields(type(self)).t_samp.default.factory()

    def _update_saved_vars(self, t, i, t_samp):
        if self.probe.save_history:
            self.i = np.concatenate([self.i, i])
            self.t = unit_safe_cat([self.t, t])
            t_samp_rep = np.full_like(t, t_samp)
            self.t_samp = unit_safe_cat([self.t_samp, t_samp_rep])

    @property
    @abstractmethod
    def n(self):
        """Number of spike sources: channels for :class:`MultiUnitActivity` or sorted neurons
        for :class:`SortedSpiking`."""
        pass

    @property
    def n_channels(self) -> int:
        """Number of channels on probe"""
        return self.probe.n

    @property
    def n_neurons(self) -> int:
        """Number of neurons recorded by probe"""
        return sum([np.sum(i_probe != -2) for i_probe in self.i_probe_by_ng.values()])

    @property
    def r_threshold(self, resolution: Quantity = um / 10) -> Quantity:
        """The distance from a contact at which the SNR equals the detection threshold.
        This also means 50% single-channel recall."""
        return self.r_for_snr(self.threshold_sigma, resolution=resolution)

    def r_for_recall(
        self,
        recall: float,
        resolution: Quantity = um / 10,
        upper_limit: Quantity = None,
    ) -> Quantity:
        """The distance from a contact at which the single-channel recall
        (detection probability) equals the specified value."""
        if upper_limit is None:
            upper_limit = 10 * self.r_noise_floor
        rr = np.arange(0, upper_limit / um, resolution / um) * um
        recalls = self.recall_by_distance(rr)
        try:
            return rr[recalls <= recall][0]
        except IndexError:
            return self.r_for_recall(
                recall, resolution=resolution, upper_limit=upper_limit * 2
            )

    def r_for_snr(
        self, snr: float, resolution: Quantity = um / 10, upper_limit: Quantity = None
    ) -> Quantity:
        """The distance from a contact at which the SNR equals the specified value."""
        if upper_limit is None:
            upper_limit = 10 * self.r_noise_floor
        rr = np.arange(0, upper_limit / um, resolution / um) * um
        try:
            return rr[self.snr_by_distance(rr) <= snr][0]
        except IndexError:
            return self.r_for_snr(
                snr, resolution=resolution, upper_limit=upper_limit * 2
            )

    def recall_by_snr(self, snr: float) -> float:
        """Probability of detecting a spike at distance r from the neuron
        as a function of SNR."""
        # 1 - P(spike not detected)
        mu = snr
        sigma = np.sqrt(1 + (mu * self.spike_amplitude_cv) ** 2)
        return norm.sf(self.threshold_sigma, loc=mu, scale=sigma)

    def recall_by_distance(self, r: Quantity) -> float:
        """Probability of detecting a spike at distance r from the neuron
        as a function of distance."""
        return self.recall_by_snr(self.snr_by_distance(r))

    def snr_by_distance(self, r: Quantity) -> float:
        """The mean extracellular action potential amplitude as a function of distance
        from the neuron, in units of background noise standard deviation."""
        return self.eap_decay_fn(r + self.r0) / self.eap_decay_fn(
            self.r0 + self.r_noise_floor
        )

    def connect_to_neuron_group(self, neuron_group: NeuronGroup, **kwparams):
        """Configure signal to record from specified neuron group

        Parameters
        ----------
        neuron_group : NeuronGroup
            Neuron group to record from
        """
        super(Spiking, self).connect_to_neuron_group(neuron_group, **kwparams)
        if neuron_group in self.i_probe_by_ng:
            raise ValueError(
                f"Spiking signal {self.name} already connected to NeuronGroup {neuron_group.name}"
            )
        # could support separate detection probabilities per group using kwparams
        # n_neurons X n_channels X 3
        dist2 = np.zeros((len(neuron_group), self.n_channels))
        for dim in ["x", "y", "z"]:
            dim_ng, dim_probe = np.meshgrid(
                getattr(neuron_group, dim),
                getattr(self.probe, f"{dim}s"),
                indexing="ij",
            )
            # proactively strip units to avoid numpy maybe doing so
            dist2 += (dim_ng / mm - dim_probe / mm) ** 2
        distances = np.sqrt(dist2) * mm
        snr = self.snr_by_distance(distances)
        # 1 from baseline noise, mu * cv from spike amp variability
        sigma_eap = np.sqrt(1 + (snr * self.spike_amplitude_cv) ** 2)
        probs_miss = norm.cdf(self.threshold_sigma, loc=snr, scale=sigma_eap)
        assert probs_miss.shape == (len(neuron_group), self.n_channels)
        multi_channel_recall = 1 - np.prod(probs_miss, axis=1)
        where_to_keep = multi_channel_recall >= self.recording_recall_cutoff
        i2keep = np.nonzero(where_to_keep)[0]
        # spike sorters don't sort by multi-channel recall; we only do this
        # for computational efficiency, to not get super distant neurons

        n2keep = np.sum(where_to_keep)
        if n2keep == 0:
            warnings.warn(
                f"NeuronGroup {neuron_group.name} has no neurons with multi-channel recall >= {self.recording_recall_cutoff}."
                " Skipping this group."
            )
            return

        # create monitor
        mon = SpikeMonitor(neuron_group)
        self._monitors.append(mon)
        self.brian_objects.add(mon)
        self._mon_spikes_already_seen.append(0)

        # update mapping from neuron group to probe index
        i_probe_start = self.n_neurons
        # -2 means not recorded
        new_i_probe = np.arange(i_probe_start, i_probe_start + n2keep)
        i_probe_for_ng = np.full(neuron_group.N, -2, dtype=int)
        i_probe_for_ng[where_to_keep] = new_i_probe
        self.i_probe_by_ng[neuron_group] = i_probe_for_ng

        # update mapping from probe index to neuron group & index
        assert len(self.i_ng_by_i_probe) == i_probe_start
        self.i_ng_by_i_probe.extend(list(zip([neuron_group] * n2keep, i2keep)))

        if not self._prev_t:
            self._prev_t = self.probe.sim.network.t

        # store neuron-channel mean and stdev of EAPs to use in subclasses
        if self._mu_eap is None:
            self._mu_eap = snr[where_to_keep]
        else:
            self._mu_eap = np.concatenate((self._mu_eap, snr[where_to_keep]), axis=0)

        return snr[where_to_keep].max(axis=1), new_i_probe

    @abstractmethod
    def get_state(
        self,
    ) -> tuple[Int[np.ndarray, "n_spikes"], Quantity, Int[np.ndarray, "{self.n}"]]:
        """Return spikes since method was last called (i, t, y)

        Returns
        -------
        tuple[Int[np.ndarray, "n_spikes"], Quantity, Int[np.ndarray, "{self.n}"]]
            (i, t, y) where i is channel (for multi-unit) or neuron (for sorted) spike
            indices, t is spike times, and y is a spike count vector suitable for control-
            theoretic uses---i.e., a 0 for every channel/neuron that hasn't spiked and a 1
            for a single spike.
        """
        pass

    def _get_new_spikes(self) -> Tuple[Int[np.ndarray, "n_spikes"], Quantity]:
        i_probe = np.array([], dtype=int)
        t = ms * np.array([], dtype=float)
        for j in range(len(self._monitors)):
            mon = self._monitors[j]
            spikes_already_seen = self._mon_spikes_already_seen[j]
            i_ng = mon.i[spikes_already_seen:]  # can contain spikes we don't care about
            # filter out spikes we don't care about
            i_probe_unfilt = self.i_probe_by_ng[mon.source][i_ng]
            i2keep = i_probe_unfilt != -2
            i_probe = np.concatenate((i_probe, i_probe_unfilt[i2keep]))
            t = unit_safe_cat([t, mon.t[spikes_already_seen:][i2keep]])

            self._mon_spikes_already_seen[j] = mon.num_spikes

        return i_probe, t

    @staticmethod
    @cache
    def _prep_noise(dt_s: float) -> np.ndarray:
        lowcut_Hz = 300
        fs_Hz = 1 / dt_s
        highcut_Hz = min(3000, 0.45 * fs_Hz)
        sos = signal.butter(
            6, [lowcut_Hz, highcut_Hz], fs=fs_Hz, btype="band", output="sos"
        )
        w, h = signal.sosfreqz(sos, 2**14, fs=fs_Hz)
        enbw = np.trapezoid(np.abs(h) ** 2, w)
        assert np.isclose(enbw, highcut_Hz - lowcut_Hz, rtol=0.01), (
            f"{enbw} != {highcut_Hz - lowcut_Hz}"
        )
        nyq = 0.5 * fs_Hz
        pre_filter_factor = np.sqrt(nyq / enbw)

        # inline test:
        white_noise = rng.standard_normal((400, 32))
        # scale noise so RMS after filtering is 1
        white_noise *= pre_filter_factor
        filtered_noise = signal.sosfilt(sos, white_noise, axis=0)
        assert np.isclose(np.std(filtered_noise), 1, rtol=0.05), (
            f"Filtered noise RMS {np.std(filtered_noise)} != 1"
        )

        return sos, pre_filter_factor

    @staticmethod
    @cache
    def _cannot_simulate_noise(dt_s: float) -> bool:
        fs_Hz = 1 / dt_s
        cannot_simulate_noise = fs_Hz / 2 < 300
        if cannot_simulate_noise:
            warnings.warn(
                f"Sampling frequency {fs_Hz} Hz is too low to simulate spiking band noise"
            )
        return cannot_simulate_noise

    def _generate_noise(self) -> tuple[Float[np.ndarray, "n_t n_channels"], Quantity]:
        """generate noise in spiking band"""
        dt = b2.defaultclock.dt
        n_t = int(round((self.probe.sim.network.t - self._prev_t) / dt))
        t_window = np.arange(n_t) * dt + self._prev_t

        if self._cannot_simulate_noise(dt / b2.second):
            return np.zeros((n_t, self.n_channels)), t_window

        sos, pre_filter_factor = self._prep_noise(dt / b2.second)
        if n_t <= 0:
            return np.zeros((0, self.n_channels)), [] * ms
        # generate white noise
        white_noise = rng.standard_normal((n_t, self.n_channels))
        # scale noise so RMS after filtering is 1
        white_noise *= pre_filter_factor
        if self._prev_zi is None:
            self._prev_zi = np.zeros((sos.shape[0], 2, self.n_channels))
        noise_filt, zi = signal.sosfilt(sos, white_noise, axis=0, zi=self._prev_zi)
        # I assume we won't have spikes at the current timestep here
        # since Cleo's NetworkOperation is scheduled for start of timestep

        self._prev_zi = zi
        return noise_filt, t_window

    def _noisily_get_true_tcs(
        self, i_probe, t
    ) -> Tuple[
        Quantity,
        Int[np.ndarray, "n_tcs"],
        Int[np.ndarray, "n_tcs"],
        Float[np.ndarray, "n_tcs"],
        Float[np.ndarray, "n_t_window {self.n_channels}"],
        Quantity,
    ]:
        """"""
        n_spks = len(i_probe)
        noise, t_noise = self._generate_noise()
        # mu and sigma arrays: n_nrns x n_channels
        mu_eap_for_spikes = self._mu_eap[i_probe]
        sigma_spike_amps = mu_eap_for_spikes * self.spike_amplitude_cv

        # add noise at right timesteps
        noise_at_spks = noise[
            ((t - self._prev_t) / b2.defaultclock.dt).round().astype(int)
        ]

        assert (
            mu_eap_for_spikes.shape
            == sigma_spike_amps.shape
            == noise_at_spks.shape
            == (n_spks, self.n_channels)
        )
        amps = (
            rng.standard_normal((n_spks, 1)) * sigma_spike_amps
            + mu_eap_for_spikes
            + noise_at_spks
        )

        # ⬇ nonzero gives row, column indices of each nonzero element
        i_spk_tcs, i_chan_tcs = (amps > self.threshold_sigma).nonzero()
        i_probe_tcs = i_probe[i_spk_tcs]
        t_tcs = t[i_spk_tcs]
        amp_tcs = amps[i_spk_tcs, i_chan_tcs]
        self._prev_t = self.probe.sim.network.t

        return t_tcs, i_probe_tcs, i_chan_tcs, amp_tcs, noise, t_noise

    @staticmethod
    @cache
    def _max_collision_interval(dt_ms, collision_prob_fn):
        intervals = np.arange(10 / dt_ms) * dt_ms * ms
        # allows for function to return scale (like 0)
        coll_probs = np.broadcast_to(
            collision_prob_fn(intervals), intervals.shape
        ).astype(float)
        i = np.searchsorted(-coll_probs, -1e-3)
        if i == len(intervals):
            warnings.warn(
                "collision_prob_fn(10 ms) > 1e-3. "
                "Will not look for collisions over 10 ms in the past."
            )
            return 10 * ms
        else:
            return intervals[i]

    def _sample_collisions(self, t, i_chan, amps) -> Bool[np.ndarray, "n_spikes"]:
        """Filter out spikes that are too close together in time on the same channel.
        For simplicity, the first spike is kept, or the largest if simultaneous.

        Note this operates on candidate threshold crossings, not called spikes.
        This is mainly for computational efficiency, so we don't have to iterate."""
        assert np.all(np.diff(t) >= 0), "should be time-sorted"

        # need to combine with previous t, i_chan, amps
        try:
            where_window_starts = len(self._prev_t_tcs)
            t = unit_safe_cat([self._prev_t_tcs, t])
            i_chan = np.concatenate([self._prev_i_chan_tcs, i_chan])
            amps = np.concatenate([self._prev_amp_tcs, amps])
        except AttributeError:
            where_window_starts = 0

        # rows=spike 2, cols=spike 1
        t_diff = t[:, None] - t[None, :]
        amp_diff = amps[:, None] - amps[None, :]
        same_chan = i_chan[:, None] == i_chan[None, :]

        collision_prob = self.collision_prob_fn(t_diff) * same_chan
        # remove self-pairs
        np.fill_diagonal(collision_prob, 0)
        # only consider same-channel pairs, and only consider where t_spk2 >= t_spk1
        collision_prob *= t_diff >= 0
        # should be roughly lower triangular at this point if spikes are ordered by time
        # for simultaneous spikes, make sure biggest amplitude wins
        # by removing simultaneous pairs where the second is bigger
        collision_prob[(t_diff == 0) & (amp_diff > 0)] = 0

        which_collided = np.any(
            rng.uniform(size=collision_prob.shape) < collision_prob, axis=1
        )

        # save t, i_chan, amps for next call
        # earliest time needed from current time (more than enough for next sample)
        t_needed = self.probe.sim.network.t - self._max_collision_interval(
            b2.defaultclock.dt / ms, self.collision_prob_fn
        )
        # TODO: use searchsorted elsewhere
        i_oldest_needed = max(np.searchsorted(t, t_needed) - 1, 0)
        self._prev_t_tcs = t[i_oldest_needed:]
        self._prev_i_chan_tcs = i_chan[i_oldest_needed:]
        self._prev_amp_tcs = amps[i_oldest_needed:]

        return which_collided[where_window_starts:]

    def reset(self, **kwargs) -> None:
        # crucial that this be called after network restore
        # since that would reset monitors
        for j, mon in enumerate(self._monitors):
            self._mon_spikes_already_seen[j] = mon.num_spikes
        self._init_saved_vars()

    def to_neo(self) -> neo.Group:
        group = neo.Group(allowed_types=[neo.SpikeTrain])
        for i in set(self.i):
            st = neo.SpikeTrain(
                times=self.t[self.i == i] / ms * pq.ms,
                t_stop=self.probe.sim.network.t / ms * pq.ms,
            )
            st.annotate(i=int(i))
            group.add(st)

        group.annotate(export_datetime=datetime.now())
        group.name = f"{self.probe.name}.{self.name}"
        group.description = f"Exported from Cleo {self.__class__.__name__} object"
        return group


@define(eq=False)
class MultiUnitActivity(Spiking):
    """Detects (unsorted) spikes per channel."""

    collision_prob_fn: Callable[[Quantity], float] = lambda t: t < 1 * ms

    @property
    def n(self):
        return self.probe.n

    def get_state(
        self,
    ) -> tuple[
        Int[np.ndarray, "n_spikes"], Quantity, Int[np.ndarray, "{self.n_channels}"]
    ]:
        # inherit docstring
        t_samp = self.probe.sim.network.t
        i_probe, t = self._get_new_spikes()
        t_tcs, _, i_chan_tcs, amp_tcs, noise, t_noise = self._noisily_get_true_tcs(
            i_probe, t
        )
        # get false positives from noise
        if self.simulate_false_positives:
            i_t_fps, i_chan_fps = (noise > self.threshold_sigma).nonzero()
            t_fps = t_noise[i_t_fps]
            amp_fps = noise[i_t_fps, i_chan_fps]
        else:
            t_fps, i_chan_fps, amp_fps = [] * ms, [], []

        i_chan = np.concatenate([i_chan_tcs, i_chan_fps])
        t = unit_safe_cat([t_tcs, t_fps])
        amps = np.concatenate([amp_tcs, amp_fps])
        # sort by time
        sort_idx = np.argsort(t)
        t = t[sort_idx]
        i_chan = i_chan[sort_idx]
        amps = amps[sort_idx]

        # sample collisions
        which_collided = self._sample_collisions(t, i_chan, amps)
        t_detected = t[~which_collided]
        i_chan_detected = i_chan[~which_collided]

        y = np.bincount(i_chan_detected.astype(int))
        # include 0s for upper indices not seen:
        y = np.concatenate([y, np.zeros(self.n_channels - len(y))])
        self._update_saved_vars(t_detected, i_chan_detected, t_samp)
        return i_chan_detected, t_detected, y

    def to_neo(self) -> neo.Group:
        group = super(MultiUnitActivity, self).to_neo()
        for st in group.spiketrains:
            i = int(st.annotations["i"])
            st.annotate(
                i_channel=i,
                x_contact=self.probe.coords[i, 0] / mm * pq.mm,
                y_contact=self.probe.coords[i, 1] / mm * pq.mm,
                z_contact=self.probe.coords[i, 2] / mm * pq.mm,
            )
        return group


@define(eq=False)
class SortedSpiking(Spiking):
    """Detect spikes identified by neuron indices.

    The indices used by the probe do not correspond to those
    coming from neuron groups, since the probe must consider
    multiple potential groups and within a group ignores those
    neurons that are too far away to be easily detected."""

    snr_cutoff: float = 6
    """The signal-to-noise ratio a unit must have for its spikes to be reported.
    SNR is defined as the mean spike amplitude divided by the standard deviation of
    the background noise for the peak (closest) channel.
    
    Should be higher than :attr:`~Spiking.threshold_sigma`.
    Spikes from units with SNR < snr_cutoff still factor into collision sampling and
    are reported as unsorted (index -1), essentially "multi-unit activity"."""
    collision_prob_fn: Callable[[Quantity], float] = lambda t: 0.2 * np.exp(
        -t / (0.3 * ms)
    )

    @property
    def n(self):
        return self.n_sorted

    @property
    def n_sorted(self):
        """Number of sorted neurons"""
        return len(self._i_probe_by_i_sorted)

    def i_sorted_by_ng(self, ng: NeuronGroup) -> Int[np.ndarray, "{ng.N}"]:
        """Get the sorted indices for a given neuron group.

        -1 means recorded, but not sorted. -2 means not recorded."""
        i_probe = self.i_probe_by_ng[ng][ng.i]
        i_sorted = self._i_sorted_by_i_probe[i_probe]
        # pass through the -2s so they don't index i_sorted
        i_sorted[i_probe == -2] = -2
        return i_sorted

    @property
    def i_ng_by_i_sorted(self) -> list[tuple[NeuronGroup, int]]:
        """Get a list of (ng, i_ng) tuples for all sorted neurons, in order.

        That is, this maps from sorted indices back to the original neuron group and
        indices."""
        i_probe = self._i_probe_by_i_sorted[np.arange(self.n_sorted)]
        return [self.i_ng_by_i_probe[i] for i in i_probe]

    _i_sorted_by_i_probe: Int[np.ndarray, "{self.n_neurons}"] = field(
        init=False, factory=lambda: np.zeros(0, dtype=int), repr=False
    )
    _i_probe_by_i_sorted: Int[np.ndarray, "{self.n_sorted}"] = field(
        init=False, factory=lambda: np.zeros(0, dtype=int), repr=False
    )

    @property
    def r_cutoff(self, resolution: Quantity = um / 10) -> Quantity:
        """The distance from a contact at which the SNR is high enough for a neuron
        to be included."""
        return self.r_for_snr(self.snr_cutoff, resolution=resolution)

    @property
    def sorted_units_snr(self) -> Float[np.ndarray, "{self.n_sorted}"]:
        """The SNR for each sorted neuron, in order."""
        return self._mu_eap[self._i_probe_by_i_sorted]

    def connect_to_neuron_group(self, neuron_group, **kwparams):
        snr, i_probe = super().connect_to_neuron_group(neuron_group, **kwparams)
        n_recorded_ng = len(snr)
        # filter by SNR (measured on peak channel)
        above_cutoff = snr >= self.snr_cutoff
        n_above_cutoff = np.sum(above_cutoff)
        n_prev_sorted = self.n_sorted
        i_srt_new_range = np.arange(self.n_sorted, n_prev_sorted + n_above_cutoff)

        # update map from i_probe to i_sorted
        # -1 means not reported
        i_srt_by_i_probe_for_ng = np.full(n_recorded_ng, -1, dtype=int)
        i_srt_by_i_probe_for_ng[above_cutoff] = i_srt_new_range
        self._i_sorted_by_i_probe = np.concatenate(
            [self._i_sorted_by_i_probe, i_srt_by_i_probe_for_ng]
        )

        # update map from i_sorted to i_probe
        i_probe_by_i_sorted_for_ng = i_probe[above_cutoff]
        assert len(i_probe_by_i_sorted_for_ng) == n_above_cutoff
        self._i_probe_by_i_sorted = np.concatenate(
            [self._i_probe_by_i_sorted, i_probe_by_i_sorted_for_ng]
        )
        assert n_prev_sorted + n_above_cutoff == self.n_sorted

    def get_state(
        self,
    ) -> tuple[Int[np.ndarray, "n_spikes"], Quantity, Int[np.ndarray, "{self.n}"]]:
        # inherit docstring

        t_samp = self.probe.sim.network.t
        i_probe, t = self._get_new_spikes()
        t_tcs, i_probe_tcs, i_chan_tcs, amp_tcs, noise, t_noise = (
            self._noisily_get_true_tcs(i_probe, t)
        )

        # get false positives from noise
        if self.simulate_false_positives:
            i_t_fps, i_chan_fps = (noise > self.threshold_sigma).nonzero()
            t_fps = t_noise[i_t_fps]
            amp_fps = noise[i_t_fps, i_chan_fps]
        else:
            t_fps, i_chan_fps, amp_fps = [] * ms, [], []

        i_chan = np.concatenate([i_chan_tcs, i_chan_fps])
        t = unit_safe_cat([t_tcs, t_fps])
        amps = np.concatenate([amp_tcs, amp_fps])
        i_probe = np.concatenate([i_probe_tcs, np.full_like(i_chan_fps, -3)]).astype(
            int
        )
        # sort by time
        sort_idx = np.argsort(t)
        t = t[sort_idx]
        i_chan = i_chan[sort_idx]
        amps = amps[sort_idx]
        i_probe = i_probe[sort_idx]

        which_collided = self._sample_collisions(t, i_chan, amps)
        # filter out false positives
        t_detected = t[~which_collided & (i_probe != -3)]
        i_probe_detected = i_probe[~which_collided & (i_probe != -3)]

        # remove repeat t, i_nrn spikes (spikes detected on >1 channel)
        _, spk_detected_any_channel = np.unique(
            np.array([t_detected / ms, i_probe_detected]), axis=1, return_index=True
        )
        # get spikes detected on any channel
        i_probe_detected = i_probe_detected[spk_detected_any_channel]
        t_detected = t_detected[spk_detected_any_channel]
        # convert to sorted indices, including getting -1s for unsorted
        i_sorted_detected = self._i_sorted_by_i_probe[i_probe_detected]
        # filter out -1s
        to_keep = i_sorted_detected != -1
        i_srt_dtct_filt = i_sorted_detected[to_keep]
        t_dtct_filt = t_detected[to_keep]

        y = np.bincount(i_srt_dtct_filt.astype(int))
        # include 0s for upper indices not seen:
        y = np.concatenate([y, np.zeros(self.n_sorted - len(y))])
        self._update_saved_vars(t_dtct_filt, i_srt_dtct_filt, t_samp)
        return i_srt_dtct_filt, t_dtct_filt, y



from __future__ import annotations

import warnings
from typing import Callable, Tuple

import matplotlib.pyplot as plt
from attrs import define, field
from brian2 import NeuronGroup, Quantity, mm, nmeter, np
from scipy.interpolate import (
    Akima1DInterpolator,
    CubicSpline,
    PchipInterpolator,
    interp1d,
)

from cleo.coords import assign_xyz
from cleo.utilities import brian_safe_name, wavelength_to_rgb


def linear_interpolator(lambdas_nm, epsilons, lambda_new_nm):
    # return np.interp(lambda_new_nm, lambdas_nm, epsilons)
    return interp1d(lambdas_nm, epsilons, fill_value="extrapolate")(lambda_new_nm)


def cubic_interpolator(lambdas_nm, epsilons, lambda_new_nm):
    return CubicSpline(lambdas_nm, epsilons)(lambda_new_nm)


def pchip_interpolator(lambdas_nm, epsilons, lambda_new_nm):
    return PchipInterpolator(lambdas_nm, epsilons)(lambda_new_nm)


def makima_interpolator(lambdas_nm, epsilons, lambda_new_nm):
    intrp = Akima1DInterpolator(lambdas_nm, np.log(epsilons))
    intrp.extrapolate = True
    intrp.method = "makima"
    return intrp(lambda_new_nm)


def _log_(fn, lambdas_nm, epsilons, lambda_new_nm):
    assert isinstance(epsilons, np.ndarray)
    epsilons[epsilons == 0] = 1e-10
    return np.exp(fn(lambdas_nm, np.log(epsilons), lambda_new_nm))


def log_linear_interpolator(lambdas_nm, epsilons, lambda_new_nm):
    return _log_(linear_interpolator, lambdas_nm, epsilons, lambda_new_nm)


def log_pchip_interpolator(lambdas_nm, epsilons, lambda_new_nm):
    return _log_(pchip_interpolator, lambdas_nm, epsilons, lambda_new_nm)


def log_makima_interpolator(lambdas_nm, epsilons, lambda_new_nm):
    return _log_(makima_interpolator, lambdas_nm, epsilons, lambda_new_nm)


# hacky MRO stuff...multiple inheritance only works because slots=False,
# and must be placed *before* SynapseDevice to work right
@define(eq=False, slots=False)
class LightDependent:
    """Mix-in class for opsin and light-dependent indicator.
    Light-dependent devices are connected to light sources (and vice-versa)
    on injection via the registry.

    We approximate dynamics under multiple wavelengths using a weighted sum
    of photon fluxes, where the ε factor indicates the activation
    relative to the peak-sensitivity wavelength for a equivalent power, which
    most papers report. When they report the action spectrum for equivalent
    photon flux instead (see Mager et al, 2018), use :func:`equal_photon_flux_spectrum`.
    This weighted sum is an approximation of a nonlinear
    peak-non-peak wavelength relation; see ``notebooks/multi_wavelength_model.ipynb``
    for details."""

    spectrum: list[tuple[float, float]] = field()
    """List of (wavelength, epsilon) tuples representing the action (opsin) or
    excitation (indicator) spectrum."""

    @spectrum.default
    def _default_spectrum(self):
        warnings.warn(
            f"No spectrum provided for light-dependent device {self.name}."
            " Assuming ε = 1 for all λ."
        )
        return [(-1e10, 1), (1e10, 1)]

    spectrum_interpolator: Callable = field(default=log_pchip_interpolator, repr=False)
    """Function of signature (lambdas_nm, epsilons, lambda_new_nm) that interpolates
    the action spectrum data and returns :math:`\\varepsilon \\in [0,1]` for the new
    wavelength."""
    extrapolate: bool = False
    """Whether or not to attempt to extrapolate (using :attr:`spectrum_interpolator`)
    outside of the provided excitation/action spectrum."""

    @property
    def light_agg_ngs(self) -> dict[str, NeuronGroup]:
        """Returns the "neurons" that aggregate light for this device.
        Dict of form {target_ng.name: light_agg_ng}."""
        return self.source_ngs

    def _get_source_for_synapse(
        self, target_ng: NeuronGroup, i_targets: list[int]
    ) -> Tuple[NeuronGroup, list[int]]:
        # create light aggregator neurons
        light_agg_ng = NeuronGroup(
            len(i_targets),
            model="""
            phi : 1/second/meter**2
            Irr : watt/meter**2
            """,
            name=f"light_agg_{brian_safe_name(self.name)}_{target_ng.name}",
        )
        assign_xyz(
            light_agg_ng,
            target_ng.x[i_targets] / mm,
            target_ng.y[i_targets] / mm,
            target_ng.z[i_targets] / mm,
            unit=mm,
        )
        return light_agg_ng, list(range(len(i_targets)))

    def epsilon(self, lambda_new: Quantity) -> float:
        """Returns the :math:`\\varepsilon` value for a given lambda (including units)
        representing the relative sensitivity of the opsin to that wavelength."""
        lam_eps_array = np.array(self.spectrum)
        lambdas, epsilons = lam_eps_array[lam_eps_array[:, 0].argsort()].T
        lambda_new /= nmeter
        eps_new = self.spectrum_interpolator(lambdas, epsilons, lambda_new)

        # out of data range
        if lambda_new < min(lambdas) or lambda_new > max(lambdas):
            if not self.extrapolate:
                warnings.warn(
                    f"λ = {lambda_new} nm is outside the range of the action spectrum data"
                    f" for {self.name} and extrapolate=False. Assuming ε = 0."
                )
                return 0
            elif self.extrapolate:
                warnings.warn(
                    f"λ = {lambda_new} nm is outside the range of the action spectrum data"
                    f" for {self.name}. Extrapolating: ε = {eps_new:.3f}."
                )
        if eps_new < 0:
            warnings.warn(f"ε = {eps_new} < 0 for {self.name}. Setting ε = 0.")
            eps_new = 0
        if eps_new > 1:
            warnings.warn(f"ε = {eps_new} > 1 for {self.name}. Setting ε = 1.")
            eps_new = 1
        return eps_new


def equal_photon_flux_spectrum(
    spectrum: list[tuple[float, float]],
) -> list[tuple[float, float]]:
    """Converts an equival photon flux spectrum to an equal power density spectrum."""
    spectrum = np.array(spectrum)
    lambdas = spectrum[:, 0]
    eps_phi = spectrum[:, 1]
    eps_Irr = eps_phi / lambdas
    eps_Irr /= np.max(eps_Irr)
    return list(zip(lambdas, eps_Irr))


def plot_spectra(
    *ldds: LightDependent, extrapolate: bool = False, range: str = "1p"
) -> tuple[plt.Figure, plt.Axes]:
    """Plots the action/excitation spectra for multiple light-dependent devices

    Parameters
    ----------
    *ldds : LightDependent
        Device(s) to plot spectra for
    extrapolate : bool, optional
        Whether to plot extrapolated spectra, by default False
    range : str, optional
        "1p", "2p", or "1p2p", indicating What wavelengths to plot, by default "1p"

    Returns
    -------
    tuple[plt.Figure, plt.Axes]
        The Figure and Axes objects containing the plot

    Raises
    ------
    ValueError
        For an incorrect `range`
    """
    import matplotlib.pyplot as plt

    if range.lower() not in ("1p", "2p", "1p2p"):
        raise ValueError(f"range must be 1p, 2p, or 1p2p. Got {range}.")
    xlim = {
        "1p": (350, 800),
        "1p2p": (350, 1300),
        "2p": (800, 1300),
    }[range]

    if extrapolate:
        all_lambdas = [np.array(ldd.spectrum)[:, 0] for ldd in ldds]
        lambda_min = min([lambdas.min() for lambdas in all_lambdas])
        lambda_max = max([lambdas.max() for lambdas in all_lambdas])

        lambda_min = max(lambda_min, xlim[0])
        lambda_max = min(lambda_max, xlim[1])

    fig, ax = plt.subplots()
    for ldd in ldds:
        lambdas, epsilons = np.array(ldd.spectrum).T
        i_data_in_range = (lambdas > xlim[0]) & (lambdas < xlim[1])
        lambdas, epsilons = lambdas[i_data_in_range], epsilons[i_data_in_range]
        if not extrapolate:
            lambda_min = np.min(lambdas)
            lambda_max = np.max(lambdas)

            lambda_min = max(lambda_min, xlim[0])
            lambda_max = min(lambda_max, xlim[1])

        lambdas_new = np.linspace(lambda_min, lambda_max, 100)
        epsilons_new = ldd.spectrum_interpolator(lambdas, epsilons, lambdas_new)
        c_points = [wavelength_to_rgb(l) for l in lambdas * nmeter]
        c_line = wavelength_to_rgb(lambdas[np.argmax(epsilons)] * nmeter)
        ax.plot(lambdas_new, epsilons_new, c=c_line, label=ldd.name)
        ax.scatter(lambdas, epsilons, marker="o", s=50, color=c_points)

    title = (
        "Action/excitation spectra" if len(ldds) > 1 else f"Action/excitation spectrum"
    )

    ax.set(
        xlabel="λ (nm)",
        ylabel="ε",
        title=title,
        xlim=xlim,
        ylim=[0, epsilons_new.max()],
    )
    fig.legend()
    return fig, ax



"""Basic processor definitions and control/estimation functions"""
from __future__ import annotations

from abc import abstractmethod
from collections import deque
from typing import Tuple

import numpy as np
from attrs import define, field, fields
from brian2 import Quantity, ms
from jaxtyping import UInt

from cleo.base import IOProcessor
from cleo.utilities import unit_safe_append


@define
class LatencyIOProcessor(IOProcessor):
    """IOProcessor capable of delivering stimulation some time after measurement.

    Note
    ----
    It doesn't make much sense to combine parallel computation
    with "when idle" sampling, because "when idle" sampling only produces
    one sample at a time to process.
    """

    t_samp: Quantity = field(factory=lambda: [] * ms, init=False, repr=False)
    """Record of sampling times---each time :meth:`~put_state` is called."""

    sampling: str = field(default="fixed")
    """Sampling scheme: "fixed" or "when idle".
    
    "fixed" sampling means samples are taken on a fixed schedule,
    with no exceptions.
    
        

    "when idle" sampling means no samples are taken before the previous
    sample's output has been delivered. A sample is taken ASAP
    after an over-period computation: otherwise remains on schedule.
    """

    @sampling.validator
    def _validate_sampling(self, attribute, value):
        if value not in ["fixed", "when idle"]:
            raise ValueError("Invalid sampling scheme:", value)

    processing: str = field(default="parallel")
    """Processing scheme: "serial" or "parallel".

    "parallel" computes the output time by adding the delay for a sample
    onto the sample time, so if the delay is 2 ms, for example, while the
    sample period is only 1 ms, some of the processing is happening in
    parallel. Output order matches input order even if the computed
    output time for a sample is sooner than that for a previous
    sample.

    "serial" computes the output time by adding the delay for a sample
    onto the output time of the previous sample, rather than the sampling
    time. Note this may be of limited
    utility because it essentially means the *entire* round trip
    cannot be in parallel at all. More realistic is that simply
    each block or phase of computation must be serial. If anyone
    cares enough about this, it will have to be implemented in the
    future.
    """

    @processing.validator
    def _validate_processing(self, attribute, value):
        if value not in ["serial", "parallel"]:
            raise ValueError("Invalid processing scheme:", value)

    out_buffer: deque[Tuple[dict, float]] = field(factory=deque, init=False, repr=False)
    """
        "serial" computes the output time by adding the delay for a sample
        onto the output time of the previous sample, rather than the sampling
        time. Note this may be of limited
        utility because it essentially means the *entire* round trip
        cannot be in parallel at all. More realistic is that simply
        each block or phase of computation must be serial. If anyone
        cares enough about this, it will have to be implemented in the
        future.

        Note
        ----
        It doesn't make much sense to combine parallel computation
        with "when idle" sampling, because "when idle" sampling only produces
        one sample at a time to process.

        Raises
        ------
        ValueError
            For invalid `sampling` or `processing` kwargs 
        """

    def put_state(self, state_dict: dict, t_samp: Quantity):
        self.t_samp = unit_safe_append(self.t_samp, t_samp)
        out, t_out = self.process(state_dict, t_samp)
        if self.processing == "serial" and len(self.out_buffer) > 0:
            prev_t_out = self.out_buffer[-1][1]
            # add delay onto the output time of the last computation
            t_out = prev_t_out + t_out - t_samp
        self.out_buffer.append((out, t_out))
        self._needs_off_schedule_sample = False

    def get_ctrl_signals(self, t_query):
        if len(self.out_buffer) == 0:
            return {}
        next_out_signal, next_t_out = self.out_buffer[0]
        if t_query >= next_t_out:
            self.out_buffer.popleft()
            return next_out_signal
        else:
            return {}

    def _is_currently_idle(self, t_query):
        return len(self.out_buffer) == 0 or self.out_buffer[0][1] <= t_query

    def is_sampling_now(self, t_query):
        resid_ms = np.round((t_query % self.sample_period) / ms, 6)
        if self.sampling == "fixed":
            if np.isclose(resid_ms, 0) or np.isclose(
                resid_ms, np.round(self.sample_period / ms, 6)
            ):
                return True
        elif self.sampling == "when idle":
            if np.isclose(resid_ms, 0):
                if self._is_currently_idle(t_query):
                    self._needs_off_schedule_sample = False
                    return True
                else:  # if not done computing
                    self._needs_off_schedule_sample = True
                    return False
            else:
                # off-schedule, only sample if the last sampling period
                # was missed (there was an overrun)
                return self._needs_off_schedule_sample and self._is_currently_idle(
                    t_query
                )
        return False

    @abstractmethod
    def process(self, state_dict: dict, t_samp: Quantity) -> Tuple[dict, Quantity]:
        """Process network state to generate output to update stimulators.

        This is the function the user must implement to define the signal processing
        pipeline.

        Parameters
        ----------
        state_dict : dict
            {`recorder_name`: `state`} dictionary from :func:`~cleo.CLSimulator.get_state()`
        t_samp : Quantity
            The time at which the sample was taken.

        Returns
        -------
        Tuple[dict, Quantity]
            {'stim_name': `ctrl_signal`} dictionary and output time (including unit).
        """
        pass

    def _base_reset(self):
        self.t_samp = fields(type(self)).t_samp.default.factory()
        self.out_buffer = fields(type(self)).out_buffer.default.factory()
        self._needs_off_schedule_sample = False


class RecordOnlyProcessor(LatencyIOProcessor):
    """Take samples without performing any control.

    Use this if all you are doing is recording."""

    def __init__(self, sample_period, **kwargs):
        super().__init__(sample_period, **kwargs)

    def process(self, state_dict: dict, sample_time: float) -> Tuple[dict, float]:
        return ({}, sample_time)


def exp_firing_rate_estimate(
    spike_counts: UInt[np.ndarray, "num_spike_sources"],
    dt: Quantity,
    prev_rate: Quantity,
    tau: Quantity,
) -> Quantity:
    """Estimate firing rate with a recursive exponential filter.

    Parameters
    ----------
    spike_counts: np.ndarray
        n-length vector of spike counts
    dt: Quantity
        Time since last measurement (with Brian temporal unit)
    prev_rate: Quantity
        n-length vector of previously estimated firing rates
    tau: Quantity
        Time constant of exponential filter (with Brian temporal unit)

    Returns
    -------
    Quantity
        n-length vector of estimated firing rates (with Brian units)
    """
    alpha = np.exp(-dt / tau)
    return prev_rate * alpha + (1 - alpha) * spike_counts / dt


def pi_ctrl(
    measurement: float,
    reference: float,
    integ_error: float,
    dt: Quantity,
    Kp: float,
    Ki: Quantity = 0 / ms,
):
    error = reference - measurement
    integ_error += error * dt
    return Kp * error + Ki * integ_error, integ_error



# %%
import pickle
import brian2.only as b2
import numpy as np

# %%
ng = b2.NeuronGroup(10, "dv/dt = -v / (10*ms) : 1", threshold="v>1", name="neurons")
ng.v = np.random.rand(10)
syn = b2.Synapses(ng, ng, "w : 1", on_pre="v_post += w", name="synapses")
syn.connect("i != j")
syn.w = "rand()"
st_mon = b2.StateMonitor(ng, "v", record=True, name="st_mon")
spk_mon = b2.SpikeMonitor(ng, name="spk_mon")
net = b2.Network(ng, syn, st_mon, spk_mon)
net.store(filename="netstore.pkl")

# %%
ng = b2.NeuronGroup(10, "dv/dt = -v / (10*ms) : 1", threshold="v>1", name="neurons")
ng.v = np.random.rand(10)
syn = b2.Synapses(ng, ng, "w : 1", on_pre="v_post += w", name="synapses")
syn.connect("i != j")
# syn.w = "rand()"
st_mon = b2.StateMonitor(ng, "v", record=True, name="st_mon")
spk_mon = b2.SpikeMonitor(ng, name="spk_mon")
net2 = b2.Network(ng, syn, st_mon, spk_mon)
net2.restore(filename="netstore.pkl")
# %%
# net.__dir__
import dill

b2.prefs.codegen.target = "numpy"
ng = b2.NeuronGroup(
    10,
    "dv/dt = -v / (10*ms) : volt \n du/dt = -u / ms : 1",
    name="neurons",
)
ng.v = 1 * b2.mvolt
st_mon = b2.StateMonitor(ng, ["v", "u"], record=True, name="st_mon")
net = b2.Network(ng, st_mon)
net.run(2 * b2.ms)
# st_mon.get_states()
# net.store(filename="netstore.pkl")

with open("net.pkl", "wb") as f:
    dill.dump(net, f)

# %%
import pickle

with open("netstore.pkl", "rb") as f:
    store = pickle.load(f)
store

# %%
np.zeros_like(ng.v)



from cleo.light.light import (
    KoehlerBeam,
    Light,
    LightModel,
    OpticFiber,
    fiber473nm,
)
from cleo.light.light_dependence import (
    LightDependent,
    cubic_interpolator,
    equal_photon_flux_spectrum,
    linear_interpolator,
    log_linear_interpolator,
    log_makima_interpolator,
    log_pchip_interpolator,
    makima_interpolator,
    pchip_interpolator,
    plot_spectra,
)
from cleo.light.two_photon import (
    GaussianEllipsoid,
    tp_light_from_scope,
)



"""Assorted utilities for developers."""
import warnings
from collections.abc import MutableMapping

import brian2.only as b2
import neo
import quantities as pq
from brian2 import Quantity, ms, np, second
from brian2.equations.equations import (
    DIFFERENTIAL_EQUATION,
    PARAMETER,
    SUBEXPRESSION,
    Equations,
)
from brian2.groups.group import get_dtype
from matplotlib import pyplot as plt

rng: np.random.Generator = np.random.default_rng()
"""A central random number generator.

Seed is set with :func:`cleo.utilities.set_seed`"""


def set_seed(rand_seed: int):
    """Set the seed for the central random number generator

    Parameters
    ----------
    rand_seed : int
        random seed
    """
    new_rng = np.random.default_rng(rand_seed)
    assert (
        rng.bit_generator.state["bit_generator"]
        == new_rng.bit_generator.state["bit_generator"]
    ), "should be same bit generator type (default PCG64)"
    rng.bit_generator.state = new_rng.bit_generator.state


def times_are_regular(times):
    if len(times) < 2:
        return False
    return np.allclose(np.diff(times), times[1] - times[0])


def analog_signal(t, values_no_unit, units="") -> neo.core.basesignal.BaseSignal:
    if times_are_regular(t / ms):
        return neo.AnalogSignal(
            values_no_unit,
            t_start=t[0] / ms * pq.ms,
            units=units,
            sampling_period=(t[1] - t[0]) / ms * pq.ms,
        )
    else:
        return neo.IrregularlySampledSignal(
            t / ms * pq.ms,
            values_no_unit,
            units=units,
        )


def add_to_neo_segment(
    segment: neo.core.Segment, *objects: neo.core.dataobject.DataObject
):
    """Conveniently adds multiple objects to a segment.

    Taken from :class:`neo.core.group.Group`."""
    container_lookup = {
        cls_name: getattr(segment, container_name)
        for cls_name, container_name in zip(
            segment._child_objects, segment._child_containers
        )
    }

    for obj in objects:
        cls = obj.__class__
        if hasattr(cls, "proxy_for"):
            cls = cls.proxy_for
        container = container_lookup[cls.__name__]
        container.append(obj)


def normalize_coords(coords: Quantity) -> Quantity:
    """Normalize coordinates to unit vectors."""
    return coords / np.linalg.norm(coords, axis=-1, keepdims=True)


def get_orth_vectors_for_V(V):
    """For nx3 block of row vectors V, return nx3 W1, W2 orthogonal
    vector blocks"""
    V = V.reshape((-1, 3, 1))
    n = V.shape[0]
    V = np.repeat(V, 3, axis=-1)
    assert V.shape == (n, 3, 3)
    q, r = np.linalg.qr(V)
    # get two vectors orthogonal to v from QR decomp
    W1, W2 = q[..., 1], q[..., 2]
    assert W1.shape == W2.shape == (n, 3)
    return W1.squeeze(), W2.squeeze()


def xyz_from_rθz(rs, thetas, zs, xyz_start, xyz_end):
    """Convert from cylindrical to Cartesian coordinates."""
    # not using np.linalg.norm because it strips units
    m = xyz_start.reshape((-1, 3)).shape[0]
    n = len(rs)

    cyl_length = np.sqrt(np.sum((xyz_end - xyz_start) ** 2, axis=-1, keepdims=True))
    assert cyl_length.shape in [(m, 1), (1,)]
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        c = (xyz_end - xyz_start) / cyl_length  # unit vector in direction of cylinder
    # in case cyl_length is 0, producing nans
    assert c.shape in [(m, 3), (3,)]
    if c.shape == (m, 3):
        assert cyl_length.shape == (m, 1)
        c[cyl_length.ravel() == 0] = [0, 0, 1]
    elif c.shape == (3,):
        c[:] = [0, 0, 1]

    r1, r2 = get_orth_vectors_for_V(c)

    def r_unit_vecs(thetas):
        cosines = np.reshape(np.cos(thetas), (len(thetas), 1))
        sines = np.reshape(np.sin(thetas), (len(thetas), 1))
        # add axis for broadcasting so result is nx3
        cosines = np.cos(thetas)[..., np.newaxis]
        sines = np.sin(thetas)[..., np.newaxis]
        return r1.reshape((m, 1, 3)) * cosines + r2.reshape((m, 1, 3)) * sines

    coords = (
        xyz_start.reshape((m, 1, 3))
        + c.reshape((m, 1, 3)) * zs.reshape((1, n, 1))
        + rs.reshape((1, n, 1)) * r_unit_vecs(thetas)
    )
    assert coords.shape == (m, n, 3)
    return coords[..., 0], coords[..., 1], coords[..., 2]
    # coords = coords.reshape((m * n), 3)
    # return coords[:, 0], coords[:, 1], coords[:, 2]


def uniform_cylinder_rθz(n, rmax, zmax):
    """uniformly fills a cylinder with radius rmax and height zmax with n points.

    Does so by generating a Fibonacci spiral cylinder by rotating around axis
    and up and down cylinder simultaneously, using different angle steps."""
    indices = np.arange(0, n) + 0.5
    rs = rmax * np.sqrt(indices / n)
    golden_angle = np.pi * (1 + np.sqrt(5))
    thetas = golden_angle * indices
    # using sqrt(2) instead of golden ratio here so
    # the two angles don't coincide
    phis = 2 * np.pi * np.sqrt(2) * indices
    zs = zmax * (1 + np.sin(phis)) / 2

    return rs, thetas, zs


def modify_model_with_eqs(neuron_group, eqs_to_add):
    """Adapted from _create_variables() from neurongroup.py from Brian2
    source code v2.3.0.2
    """
    if type(eqs_to_add) == str:
        eqs_to_add = Equations(eqs_to_add)
    neuron_group.equations += eqs_to_add

    variables = neuron_group.variables

    dtype = {}
    if isinstance(dtype, MutableMapping):
        dtype["lastspike"] = neuron_group._clock.variables["t"].dtype

    for eq in eqs_to_add.values():
        dtype = get_dtype(eq, dtype)  # {} appears to be the default
        if eq.type in (DIFFERENTIAL_EQUATION, PARAMETER):
            if "linked" in eq.flags:
                # 'linked' cannot be combined with other flags
                if not len(eq.flags) == 1:
                    raise SyntaxError(
                        ('The "linked" flag cannot be ' "combined with other flags")
                    )
                neuron_group._linked_variables.add(eq.varname)
            else:
                constant = "constant" in eq.flags
                shared = "shared" in eq.flags
                size = 1 if shared else neuron_group._N
                variables.add_array(
                    eq.varname,
                    size=size,
                    dimensions=eq.dim,
                    dtype=dtype,
                    constant=constant,
                    scalar=shared,
                )
        elif eq.type == SUBEXPRESSION:
            neuron_group.variables.add_subexpression(
                eq.varname,
                dimensions=eq.dim,
                expr=str(eq.expr),
                dtype=dtype,
                scalar="shared" in eq.flags,
            )
        else:
            raise AssertionError("Unknown type of equation: " + eq.eq_type)

    # Add the conditional-write attribute for variables with the
    # "unless refractory" flag
    if neuron_group._refractory is not False:
        for eq in neuron_group.equations.values():
            if eq.type == DIFFERENTIAL_EQUATION and "unless refractory" in eq.flags:
                not_refractory_var = neuron_group.variables["not_refractory"]
                var = neuron_group.variables[eq.varname]
                var.set_conditional_write(not_refractory_var)

    # Stochastic variables
    for xi in eqs_to_add.stochastic_variables:
        try:
            neuron_group.variables.add_auxiliary_variable(
                xi, dimensions=(second**-0.5).dim
            )
        except KeyError as ex:
            warnings.warn(
                "Adding a stochastic variable to a neuron group that already"
                " has a variable with the same name."
            )

    # Check scalar subexpressions
    for eq in neuron_group.equations.values():
        if eq.type == SUBEXPRESSION and "shared" in eq.flags:
            var = neuron_group.variables[eq.varname]
            for identifier in var.identifiers:
                if identifier in neuron_group.variables:
                    if not neuron_group.variables[identifier].scalar:
                        raise SyntaxError(
                            (
                                "Shared subexpression %s refers "
                                "to non-shared variable %s."
                            )
                            % (eq.varname, identifier)
                        )


def wavelength_to_rgb(
    wavelength: Quantity, gamma: float = 0.8
) -> tuple[float, float, float]:
    """taken from http://www.noah.org/wiki/Wavelength_to_RGB_in_Python
    This converts a given wavelength of light to an
    approximate RGB color value. The wavelength must be given
    in nanometers in the range from 380 nm through 750 nm
    (789 THz through 400 THz).

    Based on code by Dan Bruton
    http://www.physics.sfasu.edu/astro/color/spectra.html
    """
    wavelength /= b2.nmeter
    if wavelength < 380:
        wavelength = 380.0
    if wavelength > 750:
        wavelength = 750.0
    if wavelength >= 380 and wavelength <= 440:
        attenuation = 0.3 + 0.7 * (wavelength - 380) / (440 - 380)
        R = ((-(wavelength - 440) / (440 - 380)) * attenuation) ** gamma
        G = 0.0
        B = (1.0 * attenuation) ** gamma
    elif wavelength >= 440 and wavelength <= 490:
        R = 0.0
        G = ((wavelength - 440) / (490 - 440)) ** gamma
        B = 1.0
    elif wavelength >= 490 and wavelength <= 510:
        R = 0.0
        G = 1.0
        B = (-(wavelength - 510) / (510 - 490)) ** gamma
    elif wavelength >= 510 and wavelength <= 580:
        R = ((wavelength - 510) / (580 - 510)) ** gamma
        G = 1.0
        B = 0.0
    elif wavelength >= 580 and wavelength <= 645:
        R = 1.0
        G = (-(wavelength - 645) / (645 - 580)) ** gamma
        B = 0.0
    elif wavelength >= 645 and wavelength <= 750:
        attenuation = 0.3 + 0.7 * (750 - wavelength) / (750 - 645)
        R = (1.0 * attenuation) ** gamma
        G = 0.0
        B = 0.0
    else:
        R = 0.0
        G = 0.0
        B = 0.0
    return (R, G, B)


def brian_safe_name(name: str) -> str:
    return (
        name.replace(" ", "_")
        .replace("-", "_")
        .replace(".", "_")
        .replace("(", "_")
        .replace(")", "_")
    )


def style_plots_for_docs(dark=True):
    # some hacky workaround for params not being updated until after first plot
    f = plt.figure()
    plt.plot()
    plt.close(f)

    if dark:
        plt.style.use("dark_background")
        for obj in ["figure", "axes", "savefig"]:
            plt.rc(obj, facecolor="131416")  # color of Furo dark background
    else:
        plt.style.use("default")
    plt.rc("savefig", transparent=False)
    plt.rc("axes.spines", top=False, right=False)
    plt.rc(
        "font",
        **{
            "sans-serif": "Open Sans, DejaVu Sans, Bitstream Vera Sans, Computer Modern Sans Serif, Lucida Grande, Verdana, Geneva, Lucid, Arial, Helvetica, Avant Garde, sans-serif"
        },
    )


def style_plots_for_paper(fontscale=5 / 6):
    """
    fontscale=5/6 goes from default paper font size of 9.6 down to 8
    """
    # some hacky workaround for params not being updated until after first plot
    f = plt.figure()
    plt.plot()
    plt.close(f)

    try:
        import seaborn as sns

        sns.set_context("paper", font_scale=fontscale)
    except ImportError:
        plt.style.use("seaborn-v0_8-paper")
        warnings.warn("Seaborn not found, using matplotlib style, ignoring fontscale")
    plt.rc("savefig", transparent=True, bbox="tight", dpi=300)
    plt.rc("svg", fonttype="none")
    plt.rc("axes.spines", top=False, right=False)
    plt.rc(
        "font",
        **{
            "sans-serif": "Open Sans, DejaVu Sans, Bitstream Vera Sans, Computer Modern Sans Serif, Lucida Grande, Verdana, Geneva, Lucid, Arial, Helvetica, Avant Garde, sans-serif"
        },
    )


def unit_safe_append(q1: Quantity, q2: Quantity, axis=None):
    if not b2.have_same_dimensions(q1, q2):
        raise ValueError("Dimensions (units) must match")
    if isinstance(q1, Quantity):
        assert isinstance(q2, Quantity)
        unit = q1.get_best_unit()
        return np.append(q1 / unit, q2 / unit, axis=axis) * unit
    else:
        return np.append(q1, q2, axis=axis)


def unit_safe_round(q: Quantity, decimals):
    unit = q.get_best_unit()
    return np.round(q / unit, decimals) * unit


def unit_safe_cat(quantities):
    unit = quantities[0].get_best_unit()
    return np.concatenate([q / unit for q in quantities]) * unit


def unit_safe_allclose(q1: Quantity, q2: Quantity):
    assert b2.have_same_dimensions(q1, q2)
    unit = q1.get_best_unit()
    return np.allclose(q1 / unit, q2 / unit)



"""Contains probes, convenience functions for generating electrode array coordinates,
signals, spiking, and LFP"""

from cleo.ephys.lfp import (
    LFPSignalBase,
    RWSLFPSignalBase,
    RWSLFPSignalFromPSCs,
    RWSLFPSignalFromSpikes,
    TKLFPSignal,
)
from cleo.ephys.probes import (
    Probe,
    Signal,
    linear_shank_coords,
    poly2_shank_coords,
    poly3_shank_coords,
    tetrode_shank_coords,
    tile_coords,
)
from cleo.ephys.spiking import MultiUnitActivity, SortedSpiking, Spiking



"""Contains Light device and propagation models"""

from __future__ import annotations

import datetime
import warnings
from abc import ABC, abstractmethod
from typing import Any

import matplotlib as mpl
import quantities as pq
from attrs import define, field
from brian2 import (
    NeuronGroup,
    Subgroup,
    np,
)
from brian2.units import (
    Quantity,
    mm,
    mm2,
    mwatt,
    nmeter,
    um,
)
from jaxtyping import Float
from matplotlib import colors
from matplotlib.artist import Artist
from matplotlib.collections import PathCollection

from cleo.base import CLSimulator
from cleo.coords import coords_from_xyz
from cleo.stimulators import Stimulator
from cleo.utilities import (
    analog_signal,
    normalize_coords,
    uniform_cylinder_rθz,
    wavelength_to_rgb,
    xyz_from_rθz,
)


@define
class LightModel(ABC):
    """Defines how light propagates given a source location and direction."""

    @abstractmethod
    def transmittance(
        self,
        source_coords: Quantity,
        source_direction: Float[np.ndarray, "*sources 3"],
        target_coords: Quantity,
    ) -> Float[np.ndarray, "*sources n_targets"]:
        """Output must be between 0 and 1 with shape (\\*sources, n_targets)."""
        pass

    @abstractmethod
    def viz_params(
        self,
        coords: Quantity,
        direction: Float[np.ndarray, "*sources 3"],
        T_threshold: float,
        n_points_per_source: int = None,
        **kwargs,
    ) -> tuple[Quantity, Quantity, float]:
        """Returns info needed for visualization.
        Output is ((\\*sources, n_points_per_source, 3) viz_points array, markersize, intensity_scale).

        For best-looking results, implementations should scale `markersize` and `intensity_scale`.
        """
        pass

    @property
    @abstractmethod
    def area0(self) -> Quantity:
        """The area of the light source(s), used to convert between power and irradiance."""
        pass

    def _get_rz_for_xyz(self, source_coords, source_direction, target_coords):
        """Assumes x, y, z already have units"""
        m = source_coords.reshape((-1, 3)).shape[0]
        if target_coords.ndim > 1:
            n = target_coords.shape[-2]
        elif target_coords.ndim == 1:
            n = 1
        # either shared or per-source targets
        assert target_coords.shape in [(m, n, 3), (1, n, 3), (n, 3), (3,)]

        # relative to light source(s)
        rel_coords = target_coords - source_coords.reshape((m, 1, 3))
        assert rel_coords.shape == (m, n, 3)
        # must use brian2's dot function for matrix multiply to preserve
        # units correctly.
        # zc = usf.dot(rel_coords, source_direction)  # mxn distance along cylinder axis
        #           m x n x 3    m x 1 x 3
        zc = np.sum(
            rel_coords * source_direction.reshape((-1, 1, 3)), axis=-1
        )  # mxn distance along cylinder axis
        assert zc.shape == (m, n)
        # just need length (norm) of radius vectors
        # not using np.linalg.norm because it strips units
        r = np.sqrt(
            np.sum(
                (
                    rel_coords
                    #    m x n                 m x 3
                    # --> m x n x 1             m x 1 x 3
                    - zc.reshape((m, n, 1)) * source_direction.reshape((-1, 1, 3))
                )
                ** 2,
                axis=-1,
            )
        )
        assert r.shape == (m, n)
        return r, zc


@define
class OpticFiber(LightModel):
    """Optic fiber light model from Foutz et al., 2012.

    Defaults are from paper for 473 nm wavelength."""

    R0: Quantity = 0.1 * mm
    """optical fiber radius"""
    NAfib: float = 0.37
    """optical fiber numerical aperture"""
    K: Quantity = 0.125 / mm
    """absorbance coefficient (wavelength/tissue dependent)"""
    S: Quantity = 7.37 / mm
    """scattering coefficient (wavelength/tissue dependent)"""
    ntis: float = 1.36
    """tissue index of refraction (wavelength/tissue dependent)"""

    def transmittance(
        self,
        source_coords: Quantity,
        source_dir_uvec: Float[np.ndarray, "*sources 3"],
        target_coords: Quantity,
    ) -> Float[np.ndarray, "*sources n_targets"]:
        assert np.allclose(np.linalg.norm(source_dir_uvec, axis=-1), 1)
        r, z = self._get_rz_for_xyz(source_coords, source_dir_uvec, target_coords)
        return self._Foutz12_transmittance(r, z)

    @property
    def area0(self) -> Quantity:
        return np.pi * self.R0**2

    def _Foutz12_transmittance(self, r, z, scatter=True, spread=True, gaussian=True):
        """Foutz et al. 2012 transmittance model: Gaussian cone with Kubelka-Munk propagation.

        r and z should be (n_source, n_target) arrays with units"""

        if spread:
            # divergence half-angle of cone
            theta_div = np.arcsin(self.NAfib / self.ntis)
            Rz = self.R0 + z * np.tan(
                theta_div
            )  # radius as light spreads ("apparent radius" from original code)
            C = (self.R0 / Rz) ** 2
        else:
            Rz = self.R0  # "apparent radius"
            C = 1

        if gaussian:
            G = 1 / np.sqrt(2 * np.pi) * np.exp(-2 * (r / Rz) ** 2)
        else:
            G = 1

        if scatter:
            S = self.S
            a = 1 + self.K / S
            b = np.sqrt(a**2 - 1)
            dist = np.sqrt(r**2 + z**2)
            M = b / (a * np.sinh(b * S * dist) + b * np.cosh(b * S * dist))
        else:
            M = 1

        T = G * C * M
        T[z < 0] = 0
        return T

    def viz_params(
        self,
        coords: Quantity,
        direction: Float[np.ndarray, "*sources 3"],
        T_threshold: float,
        n_points_per_source: int = 4000,
        **kwargs,
    ) -> Quantity:
        r_thresh, zc_thresh = self._find_rz_thresholds(T_threshold)
        r, theta, zc = uniform_cylinder_rθz(n_points_per_source, r_thresh, zc_thresh)

        end = coords + zc_thresh * direction
        x, y, z = xyz_from_rθz(r, theta, zc, coords, end)
        density_factor = 3
        cyl_vol = np.pi * r_thresh**2 * zc_thresh
        markersize = (cyl_vol / n_points_per_source * density_factor) ** (1 / 3)
        intensity_scale = 1.5 * (4e3 / n_points_per_source) ** (1 / 3)
        return coords_from_xyz(x, y, z), markersize, intensity_scale

    def _find_rz_thresholds(self, thresh):
        """find r and z thresholds for visualization purposes"""
        res_mm = 0.01
        zc = np.arange(20, 0, -res_mm) * mm  # ascending T
        T = self._Foutz12_transmittance(0 * mm, zc)
        try:
            zc_thresh = zc[np.searchsorted(T, thresh)]
        except IndexError:  # no points above threshold
            zc_thresh = np.max(zc)
        # look at half the z threshold for the r threshold
        r = np.arange(20, 0, -res_mm) * mm
        T = self._Foutz12_transmittance(r, zc_thresh / 2)
        try:
            r_thresh = r[np.searchsorted(T, thresh)]
        except IndexError:  # no points above threshold
            r_thresh = np.max(r)
        # multiply by 1.2 just in case
        return r_thresh * 1.2, zc_thresh


def fiber473nm(
    R0=0.1 * mm,  # optical fiber radius
    NAfib=0.37,  # optical fiber numerical aperture
    K=0.125 / mm,  # absorbance coefficient
    S=7.37 / mm,  # scattering coefficient
    ntis=1.36,  # tissue index of refraction
) -> OpticFiber:
    """Returns an :class:`OpticFiber` model with parameters for 473 nm light.

    Parameters from Foutz et al., 2012."""
    return OpticFiber(
        R0=R0,
        NAfib=NAfib,
        K=K,
        S=S,
        ntis=ntis,
    )


@define
class KoehlerBeam(LightModel):
    """Even illumination over a circular area, with no scattering."""

    radius: Quantity
    """The radius of the Köhler beam"""
    zmax: Quantity = 500 * um
    """The maximum extent of the Köhler beam, 500 μm by default
    (i.e., no thicker than necessary to go through a slice or culture)."""

    def transmittance(self, source_coords, source_dir_uvec, target_coords):
        r, z = self._get_rz_for_xyz(source_coords, source_dir_uvec, target_coords)
        T = np.ones_like(r)
        T[r > self.radius] = 0
        T[z > self.zmax] = 0
        T[z < 0] = 0
        return T

    def viz_params(
        self, coords, direction, T_threshold, n_points_per_source=4000, **kwargs
    ):
        r, theta, zc = uniform_cylinder_rθz(n_points_per_source, self.radius, self.zmax)

        end = coords + self.zmax * direction
        x, y, z = xyz_from_rθz(r, theta, zc, coords, end)
        density_factor = 2
        cyl_vol = np.pi * self.radius**2 * self.zmax
        markersize = (cyl_vol / n_points_per_source * density_factor) ** (1 / 3)
        intensity_scale = (1 / n_points_per_source) ** (1 / 3)
        return coords_from_xyz(x, y, z), markersize, intensity_scale

    @property
    def area0(self) -> Quantity:
        return np.pi * self.radius**2


def _is_power(q: Quantity) -> bool:
    return q.has_same_dimensions(mwatt)


def _is_irr(q: Quantity) -> bool:
    return q.has_same_dimensions(mwatt / mm2)


@define(eq=False)
class Light(Stimulator):
    """Delivers light to the network for photostimulation and (when implemented) imaging.

    Requires neurons to have 3D spatial coordinates already assigned.

    Visualization kwargs
    --------------------
    n_points_per_source : int, optional
        The number of points per light source used to represent light intensity in
        space. Default varies by :attr:`light_model`. Alias ``n_points``.
    T_threshold : float, optional
        The transmittance below which no points are plotted. By default 1e-3.
    intensity : float, optional
        How bright the light appears, should be between 0 and 1. By default 0.5.
    rasterized : bool, optional
        Whether to render as rasterized in vector output, True by default.
        Useful since so many points makes later rendering and editing slow.
    """

    light_model: LightModel = field(kw_only=True)
    """Defines how light is emitted. See :class:`OpticFiber` for an example."""

    coords: Quantity = field(
        default=(0, 0, 0) * mm, converter=lambda x: np.reshape(x, (-1, 3)), repr=False
    )
    """(x, y, z) coords with Brian unit specifying where to place
    the base of the light source, by default (0, 0, 0)*mm.
    Can also be an nx3 array for multiple sources.
    """

    wavelength: Quantity = field(default=473 * nmeter, kw_only=True)
    """light wavelength with unit (usually nmeter)"""

    @coords.validator
    def _check_coords(self, attribute, value):
        if len(value.shape) != 2 or value.shape[1] != 3:
            raise ValueError(
                "coords must be an n by 3 array (with unit) with x, y, and z"
                "coordinates for n contact locations."
            )

    direction: Float[np.ndarray, "*sources 3"] = field(
        default=(0, 0, 1), converter=normalize_coords
    )
    """(x, y, z) vector specifying direction in which light
    source is pointing, by default (0, 0, 1).
    
    Will be converted to unit magnitude."""

    max_value: Quantity = field(default=None, kw_only=True)
    """The maximum value (power or irradiance) the light source can emit.
    
    Usually determined by hardware in a real experiment."""

    max_value_viz: Quantity = field(default=None, kw_only=True)
    """Maximum value (power or irradiance) for visualization purposes. 
    
    i.e., the level at or above which the light appears maximally bright.
    Only relevant in video visualization.
    """

    @property
    def color(self):
        """Color of light"""
        return wavelength_to_rgb(self.wavelength)

    def transmittance(self, target_coords) -> np.ndarray:
        """Returns :attr:`light_model` transmittance given light's coords and direction."""
        return self.light_model.transmittance(
            self.coords, self.direction, target_coords
        )

    def init_for_simulator(self, sim: CLSimulator) -> None:
        sim.registry.init_register_light(self)
        self.reset()

    def connect_to_neuron_group(
        self, neuron_group: NeuronGroup, **kwparams: Any
    ) -> None:
        registry = self.sim.registry
        if self in registry.lights_for_ng.get(neuron_group, set()):
            raise ValueError(
                f"Light {self} already connected to neuron group {neuron_group}"
            )
        registry.register_light(self, neuron_group)

    @property
    def n(self):
        """Number of light sources"""
        assert len(self.coords.shape) == 2 or len(self.coords.shape) == 1
        return len(self.coords) if len(self.coords.shape) == 2 else 1

    @property
    def source(self) -> Subgroup:
        """Returns the "neuron(s)" representing the light source(s)."""
        if self.sim is None:
            return None
        return self.sim.registry.source_for_light(self)

    @property
    def irradiance(self) -> Quantity:
        """Returns history of light irradiance with units."""
        return self.values * mwatt / mm2

    @property
    def irradiance_(self) -> Quantity:
        """Returns history of light irradiance without units (/(mwatt/mm2))."""
        return self.irradiance / (mwatt / mm2)

    @property
    def power(self) -> Quantity:
        """Returns history of light power with units."""
        return self.irradiance * self.light_model.area0

    @property
    def power_(self) -> Quantity:
        """Returns history of light power without units (/mwatt)."""
        return self.power / mwatt

    def add_self_to_plot(self, ax, axis_scale_unit, **kwargs) -> list[PathCollection]:
        # show light with point field, assigning r and z coordinates
        # to all points
        # filter out points with <0.001 transmittance to make plotting faster

        # alias
        if "n_points" in kwargs and "n_points_per_source" not in kwargs:
            kwargs["n_points_per_source"] = kwargs.pop("n_points")
        T_threshold = kwargs.pop("T_threshold", 1e-3)

        viz_points, markersize, intensity_scale = self.light_model.viz_params(
            self.coords,
            self.direction,
            T_threshold,
            **kwargs,
        )
        assert viz_points.shape[0] == self.n
        assert viz_points.shape[2] == 3
        n_points_per_source = viz_points.shape[1]
        intensity = kwargs.get("intensity", 0.5 * intensity_scale)

        biggest_dim_pixels = max([ax.bbox.height, ax.bbox.width])
        dpi = 100  # default
        pt_per_in = 72
        biggest_dim_pt = biggest_dim_pixels / dpi * pt_per_in

        biggest_dim = (
            max(
                [
                    ax.get_xlim()[1] - ax.get_xlim()[0],
                    ax.get_ylim()[1] - ax.get_ylim()[0],
                    ax.get_zlim()[1] - ax.get_zlim()[0],
                ]
            )
            * axis_scale_unit
        )

        markersize_pt = markersize / biggest_dim * biggest_dim_pt
        markerarea = markersize_pt**2

        T = self.light_model.transmittance(self.coords, self.direction, viz_points)
        assert T.shape == (self.n, n_points_per_source)

        point_clouds = []
        for i in range(self.n):
            idx_to_plot = T[i] >= T_threshold
            point_cloud = ax.scatter(
                viz_points[i, idx_to_plot, 0] / axis_scale_unit,
                viz_points[i, idx_to_plot, 1] / axis_scale_unit,
                viz_points[i, idx_to_plot, 2] / axis_scale_unit,
                c=T[i, idx_to_plot],
                cmap=self._alpha_cmap_for_wavelength(intensity),
                vmin=0,
                vmax=1,
                marker="o",
                edgecolors="none",
                label=self.name,
                s=markerarea,
            )
            with warnings.catch_warnings():
                warnings.filterwarnings(
                    "ignore", message=".*Rasterization.*will be ignored.*"
                )
                # to make manageable in SVGs
                point_cloud.set_rasterized(kwargs.get("rasterized", True))
            point_clouds.append(point_cloud)

        handles = ax.get_legend().legend_handles
        c = wavelength_to_rgb(self.wavelength)
        opto_patch = mpl.patches.Patch(color=c, label=self.name)
        handles.append(opto_patch)
        ax.legend(handles=handles)

        return point_clouds

    def update_artists(
        self, artists: list[Artist], value, *args, **kwargs
    ) -> list[Artist]:
        value *= mwatt / mm2
        assert len(artists) == self.n
        if self.max_value_viz is not None:
            max_val = self.max_value_viz
        elif self.max_value is not None:
            max_val = self.max_value
        else:
            raise Exception(
                f"Light'{self.name}' needs max_value_viz "
                "or max_value set to visualize light intensity."
            )
        max_val = self._val_same_unit_as(max_val, value)

        updated_artists = []
        for point_cloud, source_value in zip(artists, value):
            prev_value = getattr(point_cloud, "_prev_value", None)
            if source_value != prev_value:
                intensity = (
                    source_value / max_val if source_value <= max_val else max_val
                )
                point_cloud.set_cmap(self._alpha_cmap_for_wavelength(intensity))
                updated_artists.append(point_cloud)
                point_cloud._prev_value = source_value

        return updated_artists

    def _preprocess_value(self, value: Quantity) -> Quantity:
        """ensures value is a 1D, n-length Quantity of proper units"""
        if isinstance(value, (int, float)) and value == 0:
            value = 0 * mwatt / mm2
        elif not isinstance(value, Quantity):
            raise ValueError(
                f"Input to light must be a Quantity. Got {type(value)} instead."
            )
        if np.shape(value) not in [(), (1,), (self.n,)]:
            raise ValueError(
                f"Input to light must be a scalar or an array of"
                f" length {self.n}. Got {value.shape} instead."
            )
        unit = value.get_best_unit()
        value = np.broadcast_to(value / unit, (self.n,)) * unit
        if not (_is_power(value) or _is_irr(value)):
            raise ValueError(
                f"Input to light must be in units of power or irradiance."
                f" Got {value.dim} instead."
            )
        return value

    def _val_same_unit_as(self, value: Quantity, target: Quantity) -> Quantity:
        if value.has_same_dimensions(target):
            return value
        elif _is_power(value) and _is_irr(target):
            return value / self.light_model.area0
        elif _is_irr(value) and _is_power(target):
            return value * self.light_model.area0
        else:
            raise ValueError(
                f"value and target must both be power or irradiance. "
                f"Got {value.dim} and {target.dim} instead."
            )

    def update(self, value: Quantity) -> None:
        """Set the light power or irradiance. The units of the input value
        determine which.

        Parameters
        ----------
        value : Quantity
            The light irradiance (mW/mm^2) or power (mW). If a scalar, it will be
            broadcast to all sources. If an array, it must be the same shape
            as the number of sources.
        """
        value = self._preprocess_value(value)
        if np.any(value < 0):
            warnings.warn(f"{self.name}: negative light value clipped to 0")
            value[value < 0] = 0
        if self.max_value is not None:
            max_val = self._val_same_unit_as(self.max_value, value)
            value[value > max_val] = max_val

        if _is_power(value):
            irr0 = value / self.light_model.area0
        else:
            assert _is_irr(value)
            irr0 = value

        # values are stored as irradiance with mW/mm2 stripped
        super(Light, self).update(irr0 / (mwatt / mm2))
        if self.source is not None:
            self.source.Irr0 = irr0

    def _alpha_cmap_for_wavelength(self, intensity):
        c = wavelength_to_rgb(self.wavelength)
        c_dimmest = (*c, 0)
        alpha_max = 0.6
        alpha_brightest = alpha_max * intensity
        c_brightest = (*c, alpha_brightest)
        return colors.LinearSegmentedColormap.from_list(
            # breaks on newer matplotlib. leave 0 and 1 implicit
            # "incr_alpha", [(0, c_dimmest), (1, c_brightest)]
            "incr_alpha",
            [c_dimmest, c_brightest],
        )

    def to_neo(self):
        if type(self.light_model).__name__ == "GaussianEllipsoid":
            values = self.power_
            unit = "mW"
        else:
            values = self.irradiance_
            unit = "mW/mm**2"
        signal = analog_signal(self.t, values, unit)
        signal.name = self.name
        signal.description = "Exported from Cleo Light device"
        signal.annotate(export_datetime=datetime.datetime.now())
        # broadcast in case of uniform direction
        _, direction = np.broadcast_arrays(self.coords, self.direction)
        signal.array_annotate(
            x=self.coords[..., 0] / mm * pq.mm,
            y=self.coords[..., 1] / mm * pq.mm,
            z=self.coords[..., 2] / mm * pq.mm,
            direction_x=direction[..., 0],
            direction_y=direction[..., 1],
            direction_z=direction[..., 2],
            i_channel=np.arange(self.n),
        )
        return signal



"""Contains LFP signals"""

from __future__ import annotations

import warnings
from collections import deque
from datetime import datetime
from itertools import chain
from math import ceil
from numbers import Number
from typing import Any, Union

import neo
import quantities as pq
import wslfp
from attrs import define, field
from brian2 import NeuronGroup, Quantity, Synapses, mm, ms, np, um
from brian2.monitors.spikemonitor import SpikeMonitor
from brian2.synapses.synapses import SynapticSubgroup
from brian2.units import Unit, uvolt
from scipy import sparse
from tklfp import TKLFP

from cleo.base import NeoExportable
from cleo.coords import coords_from_ng
from cleo.ephys.probes import Signal
from cleo.utilities import (
    analog_signal,
    unit_safe_append,
    unit_safe_cat,
)


class LFPSignalBase(Signal, NeoExportable):
    """Base class for LFP Signals.

    Injection kwargs
    ----------------
    orientation : np.ndarray, optional
        Array of shape (n_neurons, 3) representing which way is "up," that is, towards
        the surface of the cortex, for each neuron. If a single vector is given, it is
        taken to be the orientation for all neurons in the group. [0, 0, -1] is the
        default, meaning the negative z axis is "up."
    """

    t: Quantity = field(init=False, repr=False)
    """Times at which LFP is recorded, with Brian units, stored if
    :attr:`~cleo.InterfaceDevice.save_history` on :attr:`~Signal.probe`"""
    lfp: Quantity = field(init=False, repr=False)
    """Approximated LFP from every call to :meth:`get_state`.
    Shape is (n_samples, n_channels). Stored if
    :attr:`~cleo.InterfaceDevice.save_history` on :attr:`~Signal.probe`"""
    _elec_coords: Quantity = field(init=False, repr=False)
    _lfp_unit: Unit

    def _post_init_for_probe(self):
        self._elec_coords = self.probe.coords.copy()
        # need to invert z coords since cleo uses an inverted z axis and
        # tklfp and wslfp do not
        self._elec_coords[:, 2] *= -1
        self._init_saved_vars()

    def _init_saved_vars(self):
        if self.probe.save_history:
            self.t = np.empty((0,)) * ms
            self.lfp = np.empty((0, self.probe.n)) * self._lfp_unit

    def _update_saved_vars(self, t, lfp):
        if self.probe.save_history:
            lfp = np.reshape(lfp, (1, -1))
            t = np.reshape(t, (1,))
            self.t = unit_safe_append(self.t, t)
            self.lfp = unit_safe_append(self.lfp, lfp, axis=0)

    def to_neo(self) -> neo.AnalogSignal:
        # inherit docstring
        try:
            signal = analog_signal(
                self.t,
                self.lfp,
                str(self._lfp_unit),
            )
        except AttributeError:
            return
        signal.name = self.probe.name + "." + self.name
        signal.description = f"Exported from Cleo {self.__class__.__name__} object"
        signal.annotate(export_datetime=datetime.now())
        # broadcast in case of uniform direction
        signal.array_annotate(
            x=self.probe.coords[..., 0] / mm * pq.mm,
            y=self.probe.coords[..., 1] / mm * pq.mm,
            z=self.probe.coords[..., 2] / mm * pq.mm,
            i_channel=np.arange(self.probe.n),
        )
        return signal


@define(eq=False)
class TKLFPSignal(LFPSignalBase):
    """Records the Teleńczuk kernel LFP approximation.

    Requires ``tklfp_type='exc'|'inh'`` to specify cell type
    on injection.

    TKLFP is computed from spikes using the `tklfp package <https://github.com/kjohnsen/tklfp/>`_.

    Injection kwargs
    ----------------
    tklfp_type : str
        Either 'exc' or 'inh' to specify the cell type.
    """

    uLFP_threshold: Quantity = 1e-3 * uvolt
    """Threshold, in microvolts, above which the uLFP for a single spike is guaranteed
    to be considered, by default 1e-3.
    This determines the buffer length of past spikes, since the uLFP from a long-past
    spike becomes negligible and is ignored."""
    _tklfps: list[TKLFP] = field(init=False, factory=list, repr=False)
    _monitors: list[SpikeMonitor] = field(init=False, factory=list, repr=False)
    _mon_spikes_already_seen: list[int] = field(init=False, factory=list, repr=False)
    _i_buffers: list[list[np.ndarray]] = field(init=False, factory=list, repr=False)
    _t_buffers: list[list[np.ndarray]] = field(init=False, factory=list, repr=False)
    _buffer_positions: list[int] = field(init=False, factory=list, repr=False)
    _lfp_unit: Unit = uvolt

    def connect_to_neuron_group(self, neuron_group: NeuronGroup, **kwparams):
        # inherit docstring
        # prep tklfp object
        tklfp_type = kwparams.pop("tklfp_type", "not given")
        if tklfp_type not in ["exc", "inh"]:
            raise Exception(
                "tklfp_type ('exc' or 'inh') must be passed as a keyword argument to "
                "inject() when injecting an electrode with a TKLFPSignal."
            )
        orientation = kwparams.pop("orientation", np.array([[0, 0, -1]])).copy()
        orientation[:, 2] *= -1

        tklfp = TKLFP(
            neuron_group.x / mm,
            neuron_group.y / mm,
            -neuron_group.z / mm,  # invert neuron zs as well
            is_excitatory=tklfp_type == "exc",
            elec_coords_mm=self._elec_coords / mm,
            orientation=orientation,
        )

        # determine buffer length necessary for given neuron group
        # if 0 (neurons are too far away to much influence LFP)
        # then we ignore this neuron group
        buf_len = self._get_buffer_length(tklfp, **kwparams)

        if buf_len > 0:
            # prep buffers
            self._tklfps.append(tklfp)
            self._i_buffers.append([np.array([], dtype=int, ndmin=1)] * buf_len)
            self._t_buffers.append([np.array([], dtype=float, ndmin=1)] * buf_len)
            self._buffer_positions.append(0)

            # prep SpikeMonitor
            mon = SpikeMonitor(neuron_group)
            self._monitors.append(mon)
            self._mon_spikes_already_seen.append(0)
            self.brian_objects.add(mon)

    def get_state(self) -> np.ndarray:
        tot_tklfp = 0
        t_now = self.probe.sim.network.t
        # loop over neuron groups (monitors, tklfps)
        for i_mon in range(len(self._monitors)):
            self._update_spike_buffer(i_mon)
            tot_tklfp += self._tklfp_for_monitor(i_mon, t_now)
        tot_tklfp = np.reshape(tot_tklfp, (-1,)) * uvolt  # return 1D array (vector)
        self._update_saved_vars(t_now, tot_tklfp)
        return tot_tklfp

    def reset(self, **kwargs) -> None:
        for i_mon in range(len(self._monitors)):
            self._reset_buffer(i_mon)
        self._init_saved_vars()

    def _reset_buffer(self, i_mon):
        buf_len = len(self._i_buffers[i_mon])
        self._i_buffers[i_mon] = [np.array([], dtype=int, ndmin=1)] * buf_len
        self._t_buffers[i_mon] = [np.array([], dtype=float, ndmin=1)] * buf_len
        self._buffer_positions[i_mon] = 0

    def _update_spike_buffer(self, i_mon):
        mon = self._monitors[i_mon]
        n_prev = self._mon_spikes_already_seen[i_mon]
        buf_pos = self._buffer_positions[i_mon]

        # insert new spikes into buffer (overwriting anything previous)
        self._i_buffers[i_mon][buf_pos] = mon.i[n_prev:]
        self._t_buffers[i_mon][buf_pos] = mon.t[n_prev:]

        self._mon_spikes_already_seen[i_mon] = mon.num_spikes
        # update buffer position
        buf_len = len(self._i_buffers[i_mon])
        self._buffer_positions[i_mon] = (buf_pos + 1) % buf_len

    def _tklfp_for_monitor(self, i_mon, t_now):
        i = np.concatenate(self._i_buffers[i_mon])
        t = unit_safe_cat(self._t_buffers[i_mon])
        return self._tklfps[i_mon].compute(i, t / ms, [t_now / ms])

    def _get_buffer_length(self, tklfp, **kwparams):
        # need sampling period
        sample_period = kwparams.get("sample_period", None)
        if sample_period is None:
            try:
                sample_period = self.probe.sim.io_processor.sample_period
            except AttributeError:  # probably means sim doesn't have io_processor
                raise Exception(
                    "TKLFP needs to know the sampling period. Either set the simulator's "
                    f"IO processor before injecting {self.probe.name} or "
                    f"specify it on injection: .inject({self.probe.name} "
                    ", tklfp_type=..., sample_period=...)"
                )
        return np.ceil(
            tklfp.compute_min_window_ms(self.uLFP_threshold / uvolt)
            / (sample_period / ms)
        ).astype(int)


@define(eq=False)
class RWSLFPSignalBase(LFPSignalBase):
    """Base class for :class:`RWSLFPSignalFromSpikes` and :class:`RWSLFPSignalFromPSCs`.

    These signals should only be injected into neurons representing pyramidal cells with
    standard synaptic structure (see `Mazzoni, Lindén et al., 2015
    <https://journals.plos.org/ploscompbiol/article?id=10.1371/journal.pcbi.1004584>`_).

    RWSLFP is computed using the `wslfp package <https://github.com/siplab-gt/wslfp/>`_.

    ``amp_func`` and ``pop_aggregate`` can be overridden on injection.
    """

    # note these set defaults that can be overrriden on injection
    amp_func: callable = wslfp.mazzoni15_nrn
    """Function to calculate LFP amplitudes, by default ``wslfp.mazzoni15_nrn``.
    
    See `wslfp documentation <https://github.com/siplab-gt/wslfp/blob/master/notebooks/amplitude_comparison.ipynb>`_ for more info."""
    pop_aggregate: bool = False
    """Whether to aggregate currents across the population (as opposed to neurons having 
    differential contributions to LFP depending on their location). False by default."""

    wslfp_kwargs: dict = field(factory=dict)
    """Keyword arguments to pass to the WSLFP calculator, e.g., ``alpha``,
    ``tau_ampa_ms``, ``tau_gaba_ms````source_coords_are_somata``,
    ``source_dendrite_length_um``, ``amp_kwargs``, ``strict_boundaries``.
    """

    _wslfps: dict[NeuronGroup, wslfp.WSLFPCalculator] = field(
        init=False, factory=dict, repr=False
    )
    _lfp_unit: Unit = 1

    def _init_wslfp_calc(self, neuron_group: NeuronGroup, **kwparams):
        nrn_coords_um = coords_from_ng(neuron_group) / um
        nrn_coords_um[:, 2] *= -1

        orientation = np.copy(kwparams.pop("orientation", np.array([[0, 0, -1]])))
        orientation = orientation.reshape((-1, 3))
        assert np.shape(orientation)[-1] == 3
        orientation[..., 2] *= -1

        if self.pop_aggregate:
            nrn_coords_um = np.mean(nrn_coords_um, axis=0)
            orientation = np.mean(orientation, axis=0)

        wslfp_kwargs = {}
        for key in [
            "source_coords_are_somata",
            "source_dendrite_length_um",
            "amp_kwargs",
            "alpha",
            "tau_ampa_ms",
            "tau_gaba_ms",
            "strict_boundaries",
        ]:
            if key in kwparams:
                wslfp_kwargs[key] = kwparams.pop(key)
            elif key in self.wslfp_kwargs:
                wslfp_kwargs[key] = self.wslfp_kwargs[key]

        self._wslfps[neuron_group] = wslfp.from_xyz_coords(
            self._elec_coords / um,
            nrn_coords_um,
            amp_func=kwparams.pop("amp_func", self.amp_func),
            source_orientation=orientation,
            **wslfp_kwargs,
        )

    def get_state(self) -> np.ndarray:
        # round to avoid floating point errors
        t_now_ms = round(self.probe.sim.network.t / ms, 3)
        lfp = np.zeros((1, self.probe.n))
        for ng, wslfp_calc in self._wslfps.items():
            t_ampa_ms = t_now_ms - wslfp_calc.tau_ampa_ms
            t_gaba_ms = t_now_ms - wslfp_calc.tau_gaba_ms
            I_ampa, I_gaba = self._needed_current(ng, t_ampa_ms, t_gaba_ms)
            lfp += wslfp_calc.calculate(
                [t_now_ms], t_ampa_ms, I_ampa, t_gaba_ms, I_gaba, normalize=False
            )
        out = np.reshape(lfp, (-1,))  # return 1D array (vector)
        self._update_saved_vars(t_now_ms * ms, out)
        return out

    def _needed_current(
        self, ng, t_ampa_ms_ms: float, t_gaba_ms_ms: float
    ) -> np.ndarray:
        """output must have shape (n_t, n_current_sources)"""
        raise NotImplementedError

    def reset(self, **kwargs) -> None:
        self._init_saved_vars()

    def to_neo(self) -> neo.AnalogSignal:
        # inherit docstring
        try:
            signal = analog_signal(
                self.t,
                self.lfp,
            )
        except AttributeError:
            return
        signal.name = self.probe.name + "." + self.name
        signal.description = f"Exported from Cleo {self.__class__.__name__} object"
        signal.annotate(export_datetime=datetime.now())
        # broadcast in case of uniform direction
        signal.array_annotate(
            x=self.probe.coords[..., 0] / mm * pq.mm,
            y=self.probe.coords[..., 1] / mm * pq.mm,
            z=self.probe.coords[..., 2] / mm * pq.mm,
            i_channel=np.arange(self.probe.n),
        )
        return signal


@define
class SpikeToCurrentSource:
    """Stores info needed to calculate synaptic currents from spikes for a given spike source."""

    J: Union[np.ndarray, sparse.sparray]
    mon: SpikeMonitor
    biexp_kernel_params: dict[str, Any]


@define(eq=False)
class RWSLFPSignalFromSpikes(RWSLFPSignalBase):
    """Computes RWSLFP from the spikes onto pyramidal cell.

    Use this if your model does not simulate synaptic current dynamics directly.
    The parameters of this class are used to synthesize biexponential synaptic currents
    using ``wslfp.spikes_to_biexp_currents()``.
    ``ampa_syns`` and ``gaba_syns`` are lists of Synapses or SynapticSubgroup objects
    and must be passed as kwargs on injection, or else this signal will not be recorded
    for the target neurons (useful for ignoring interneurons).
    Attributes set on the signal object serve as the default, but can be overridden on injection.
    Also, in the case that parameters (e.g., ``tau1_ampa`` or ``weight``) vary by synapse,
    these can be overridden by passing a tuple of the Synapses or SynapticSubgroup object and
    a dictionary of the parameters to override.

    RWSLFP refers to the Reference Weighted Sum of synaptic currents LFP proxy from
    `Mazzoni, Lindén et al., 2015 <https://journals.plos.org/ploscompbiol/article?id=10.1371/journal.pcbi.1004584>`_.

    Injection kwargs
    ----------------
    ampa_syns : list[Synapses | SynapticSubgroup | tuple[Synapses|SynapticSubgroup, dict]]
        Synapses or SynapticSubgroup objects representing AMPA synapses (delivering excitatory currents).
        Or a tuple of the Synapses or SynapticSubgroup object and a dictionary of parameters to override.
    gaba_syns : list[Synapses | SynapticSubgroup | tuple[Synapses|SynapticSubgroup, dict]]
        Synapses or SynapticSubgroup objects representing GABA synapses (delivering inhibitory currents).
        Or a tuple of the Synapses or SynapticSubgroup object and a dictionary of parameters to override.
    weight : str | float, optional
        Name of the weight variable or parameter in the Synapses or SynapticSubgroup objects, or a float
        in the case of a single weight for all synapses. Default is 'w'.
    """

    # can override on injection: tau1|2_ampa|gaba, syn_delay, I_threshold
    tau1_ampa: Quantity = 2 * ms
    """The fall time constant of the biexponential current kernel for AMPA synapses.
    2 ms by default."""
    tau2_ampa: Quantity = 0.4 * ms
    """The time constant of subtracted part of the biexponential current kernel for AMPA synapses.
    0.4 ms by default."""
    tau1_gaba: Quantity = 5 * ms
    """The fall time constant of the biexponential current kernel for GABA synapses.
    5 ms by default."""
    tau2_gaba: Quantity = 0.25 * ms
    """The time constant of subtracted part of the biexponential current kernel for GABA synapses.
    0.25 ms by default."""
    syn_delay: Quantity = 1 * ms
    """The synaptic transmission delay, i.e., between a spike and the onset of the postsynaptic current.
    1 ms by default."""
    I_threshold: float = 1e-3
    """Threshold, as a proportion of the peak current, below which spikes' contribution
    to synaptic currents (and thus LFP) is ignored, by default 1e-3."""
    weight: str = "w"
    """Name of the weight variable or parameter in the Synapses or SynapticSubgroup objects.
    Default is 'w'."""
    _ampa_sources: dict[NeuronGroup, dict[Synapses, SpikeToCurrentSource]] = field(
        init=False, factory=dict, repr=False
    )
    _gaba_sources: dict[NeuronGroup, dict[Synapses, SpikeToCurrentSource]] = field(
        init=False, factory=dict, repr=False
    )

    def _get_weight(self, syn, weight):
        assert isinstance(weight, (Number, Quantity, str))
        if isinstance(weight, (Number, Quantity)):
            return weight

        if isinstance(syn, Synapses):
            syn_name = syn.name
            if weight in syn.variables:
                return getattr(syn, weight)
            elif weight in syn.namespace:
                return syn.namespace[weight]
        elif isinstance(syn, SynapticSubgroup):
            syn_name = syn.synapses.name
            if weight in syn.synapses.variables:
                return getattr(syn.synapses, weight)[syn._stored_indices]
            elif weight in syn.synapses.namespace:
                return syn.synapses.namespace[weight]

        raise ValueError(
            f"weight {weight} not found in {syn_name} variables or namespace"
        )

    def _create_spk2curr_source(self, syn, neuron_group, weight, biexp_kwparams):
        # need source_ng, syn_i, syn_j
        if isinstance(syn, Synapses):
            source_ng = syn.source
            syn_i, syn_j = syn.i, syn.j
        elif isinstance(syn, SynapticSubgroup):
            source_ng = syn.synapses.source
            syn_i = syn.synapses.i[syn._stored_indices]
            syn_j = syn.synapses.j[syn._stored_indices]
        else:
            raise TypeError(
                "ampa_syns and gaba_syns only take Synapses or SynapticSubgroup objects"
            )
        mon = SpikeMonitor(source_ng, record=list(np.unique(syn_i)))
        self.brian_objects.add(mon)

        J = sparse.lil_array((source_ng.N, neuron_group.N))
        w = self._get_weight(syn, weight)
        # workaround for version-specific bug where VariableView can't index sparse matrix
        syn_i = np.array(syn_i, dtype=int)
        syn_j = np.array(syn_j, dtype=int)
        J[syn_i, syn_j] = w
        J = J.tocsr()
        if self.pop_aggregate:
            J = J.sum(axis=1).reshape((-1, 1))

        return SpikeToCurrentSource(J, mon, biexp_kwparams)

    def _process_syn(self, syn, kwparams) -> tuple[Synapses, dict]:
        """handles the case when a tuple of Synapses, kwargs is passed in"""
        if isinstance(syn, (tuple, list)):
            syn, override_kwargs = syn
            kwparams = {**kwparams, **override_kwargs}
        else:
            assert isinstance(syn, (Synapses, SynapticSubgroup))

        return syn, kwparams

    def connect_to_neuron_group(self, neuron_group: NeuronGroup, **kwparams):
        # inherit docstring
        # this dict structure should allow multiple injections with no problem;
        # just the most recent injection will be used

        # prep wslfp calculator object
        if neuron_group not in self._wslfps:
            self._init_wslfp_calc(neuron_group, **kwparams)

        default_weight = kwparams.pop("weight", self.weight)
        ampa_syns = kwparams.pop("ampa_syns", [])
        gaba_syns = kwparams.pop("gaba_syns", [])

        biexp_kwparams = {}
        for key in [
            "tau1_ampa",
            "tau2_ampa",
            "tau1_gaba",
            "tau2_gaba",
            "syn_delay",
            "I_threshold",
        ]:
            biexp_kwparams[key] = kwparams.pop(key, getattr(self, key))

        if neuron_group not in self._ampa_sources:
            self._ampa_sources[neuron_group] = {}
            self._gaba_sources[neuron_group] = {}
        for ampa_syn in ampa_syns:
            ampa_syn, updated_kwparams = self._process_syn(ampa_syn, biexp_kwparams)
            weight = updated_kwparams.pop("weight", default_weight)
            self._ampa_sources[neuron_group][ampa_syn] = self._create_spk2curr_source(
                ampa_syn, neuron_group, weight, updated_kwparams
            )
        for gaba_syn in gaba_syns:
            gaba_syn, updated_kwparams = self._process_syn(gaba_syn, biexp_kwparams)
            weight = updated_kwparams.pop("weight", default_weight)
            self._gaba_sources[neuron_group][gaba_syn] = self._create_spk2curr_source(
                gaba_syn, neuron_group, weight, updated_kwparams
            )

    def _get_biexp_kwargs_from_s2cs(self, s2cs: SpikeToCurrentSource, syn_type: str):
        # check overrides, fall back on Signal-level defaults
        return {
            "tau1_ms": s2cs.biexp_kernel_params.get(
                f"tau1_{syn_type}", getattr(self, f"tau1_{syn_type}")
            )
            / ms,
            "tau2_ms": s2cs.biexp_kernel_params.get(
                f"tau2_{syn_type}", getattr(self, f"tau2_{syn_type}")
            )
            / ms,
            "syn_delay_ms": s2cs.biexp_kernel_params.get("syn_delay", self.syn_delay)
            / ms,
            "threshold": s2cs.biexp_kernel_params.get("I_threshold", self.I_threshold),
        }

    def _needed_current(self, ng, t_ampa_ms: float, t_gaba_ms: float) -> np.ndarray:
        """output must have shape (n_t, n_current_sources)"""
        n_sources = 1 if self.pop_aggregate else ng.N
        I_ampa = np.zeros((1, n_sources))
        for ampa_src in self._ampa_sources[ng].values():
            biexp_kwargs = self._get_biexp_kwargs_from_s2cs(ampa_src, "ampa")
            I_ampa += wslfp.spikes_to_biexp_currents(
                [t_ampa_ms],
                ampa_src.mon.t / ms,
                ampa_src.mon.i,
                ampa_src.J,
                **biexp_kwargs,
            )

        I_gaba = np.zeros((1, n_sources))
        for gaba_src in self._gaba_sources[ng].values():
            biexp_kwargs = self._get_biexp_kwargs_from_s2cs(gaba_src, "gaba")
            I_gaba += wslfp.spikes_to_biexp_currents(
                [t_gaba_ms],
                gaba_src.mon.t / ms,
                gaba_src.mon.i,
                gaba_src.J,
                **biexp_kwargs,
            )

        return I_ampa, I_gaba


@define(eq=False)
class RWSLFPSignalFromPSCs(RWSLFPSignalBase):
    """Computes RWSLFP from the currents onto pyramidal cells.

    Use this if your model already simulates synaptic current dynamics.
    ``Iampa_var_names`` and ``Igaba_var_names`` are lists of variable names to include
    and must be passed in as kwargs on injection or else the target neuron group will
    not contribute to this signal (desirable for interneurons).

    RWSLFP refers to the Reference Weighted Sum of synaptic currents LFP proxy from
    `Mazzoni, Lindén et al., 2015 <https://journals.plos.org/ploscompbiol/article?id=10.1371/journal.pcbi.1004584>`_.

    Injection kwargs
    ----------------
    Iampa_var_names : list[str]
        List of variable names in the neuron group representing AMPA currents.
    Igaba_var_names : list[str]
        List of variable names in the neuron group representing GABA currents.
    """

    _ampa_vars: dict[NeuronGroup, list] = field(init=False, factory=dict, repr=False)
    _gaba_vars: dict[NeuronGroup, list] = field(init=False, factory=dict, repr=False)
    _t_ampa_bufs: dict[NeuronGroup, deque[float]] = field(
        init=False, factory=dict, repr=False
    )
    _I_ampa_bufs: dict[NeuronGroup, deque[np.ndarray]] = field(
        init=False, factory=dict, repr=False
    )
    _t_gaba_bufs: dict[NeuronGroup, deque[float]] = field(
        init=False, factory=dict, repr=False
    )
    _I_gaba: dict[NeuronGroup, deque[np.ndarray]] = field(
        init=False, factory=dict, repr=False
    )

    def connect_to_neuron_group(self, neuron_group: NeuronGroup, **kwparams):
        # ^ is XOR
        if ("Iampa_var_names" in kwparams) ^ ("Igaba_var_names" in kwparams):
            raise ValueError(
                "Iampa_var_names and Igaba_var_names must be included together."
            )
        if "Iampa_var_names" not in kwparams:
            return

        I_ampa_names = kwparams.pop("Iampa_var_names")
        I_gaba_names = kwparams.pop("Igaba_var_names")
        for varname in chain(I_ampa_names, I_gaba_names):
            if not hasattr(neuron_group, varname):
                raise ValueError(
                    f"NeuronGroup {neuron_group.name} does not have a variable {varname}"
                )

        # prep wslfp calculator object
        if neuron_group in self._wslfps:
            warnings.warn(
                f"{self.name} previously connected to {neuron_group.name}."
                " Reconnecting will overwrite previous connection."
            )

        self._init_wslfp_calc(neuron_group, **kwparams)

        buf_len_ampa, buf_len_gaba = self._get_buf_lens_for_wslfp(
            self._wslfps[neuron_group]
        )
        self._t_ampa_bufs[neuron_group] = deque(maxlen=buf_len_ampa)
        self._I_ampa_bufs[neuron_group] = deque(maxlen=buf_len_ampa)
        self._t_gaba_bufs[neuron_group] = deque(maxlen=buf_len_gaba)
        self._I_gaba[neuron_group] = deque(maxlen=buf_len_gaba)

        # add underscores to get values without units
        self._ampa_vars[neuron_group] = [varname + "_" for varname in I_ampa_names]
        self._gaba_vars[neuron_group] = [varname + "_" for varname in I_gaba_names]

    def _buf_len(self, tau, dt):
        return ceil((tau + dt) / dt)

    def _get_buf_lens_for_wslfp(self, wslfp_calc, **kwparams):
        # need sampling period
        sample_period = kwparams.get("sample_period", None)
        if sample_period is None:
            try:
                sample_period = self.probe.sim.io_processor.sample_period
            except AttributeError:  # probably means sim doesn't have io_processor
                raise Exception(
                    "RSWLFPSignalFromPSCs needs to know the sampling period."
                    " Either set the simulator's IO processor before injecting"
                    f" {self.probe.name} or "
                    f" specify it on injection: .inject({self.probe.name}"
                    ", sample_period=...)"
                )
        buf_len_ampa = self._buf_len(wslfp_calc.tau_ampa_ms, sample_period / ms)
        buf_len_gaba = self._buf_len(wslfp_calc.tau_gaba_ms, sample_period / ms)
        return buf_len_ampa, buf_len_gaba

    def _curr_from_buffer(self, t_buf_ms, I_buf, t_eval_ms: float, n_sources):
        # t_eval_ms is not iterable
        empty = np.zeros((1, n_sources))
        if len(t_buf_ms) == 0 or t_buf_ms[0] > t_eval_ms or t_buf_ms[-1] < t_eval_ms:
            return empty
        # when tau is multiple of sample time, current should be collected
        # right when needed, at the left end of the buffer
        elif np.isclose(t_eval_ms, t_buf_ms[0]):
            return I_buf[0]

        # if not, should only need to interpolate between first and second positions
        # if buffer length is correct
        assert len(t_buf_ms) > 1
        if t_buf_ms[0] < t_eval_ms < t_buf_ms[1]:
            i_l, i_r = 0, 1
        else:
            warnings.warn(
                f"Time buffer is unexpected. Did a sample get skipped? "
                f"t_buf_ms={t_buf_ms}, t_eval_ms={t_eval_ms}"
            )
            i_l, i_r = None, None
            for i, t in enumerate(t_buf_ms):
                if t < t_eval_ms:
                    i_l = i
                if t >= t_eval_ms:
                    i_r = i
                    break
            if i_l is None or i_r is None or i_l >= i_r:
                warnings.warn(
                    "Signal buffer does not contain currents at needed timepoints. "
                    "Returning 0. "
                    f"t_buf_ms={t_buf_ms}, t_eval_ms={t_eval_ms}"
                )
                return empty

        I_interp = I_buf[i_l] + (I_buf[i_r] - I_buf[i_l]) * (
            t_eval_ms - t_buf_ms[i_l]
        ) / (t_buf_ms[i_r] - t_buf_ms[i_l])

        I_interp = np.reshape(I_interp, (1, n_sources))
        I_interp = np.nan_to_num(I_interp, nan=0)
        return I_interp

    def _needed_current(
        self, ng, t_ampa_ms: float, t_gaba_ms: float
    ) -> tuple[np.ndarray, np.ndarray]:
        """outputs must have shape (n_t, n_current_sources)"""
        # First add current currents to history
        # -- need to round to avoid floating point errors

        t_now_ms = round(self.probe.sim.network.t / ms, 3)
        self._t_ampa_bufs[ng].append(t_now_ms)
        self._t_gaba_bufs[ng].append(t_now_ms)

        I_ampa = np.zeros((1, ng.N))
        for I_ampa_name in self._ampa_vars[ng]:
            I_ampa += getattr(ng, I_ampa_name)

        I_gaba = np.zeros((1, ng.N))
        for I_gaba_name in self._gaba_vars[ng]:
            I_gaba += getattr(ng, I_gaba_name)

        if self.pop_aggregate:
            I_ampa = np.sum(I_ampa)
            I_gaba = np.sum(I_gaba)
        self._I_ampa_bufs[ng].append(I_ampa)
        self._I_gaba[ng].append(I_gaba)

        # Then interpolate history to get currents at the requested times
        n_sources = 1 if self.pop_aggregate else ng.N
        I_ampa = self._curr_from_buffer(
            self._t_ampa_bufs[ng], self._I_ampa_bufs[ng], t_ampa_ms, n_sources
        )
        I_gaba = self._curr_from_buffer(
            self._t_gaba_bufs[ng], self._I_gaba[ng], t_gaba_ms, n_sources
        )

        return I_ampa, I_gaba

    def reset(self, **kwargs) -> None:
        self._init_saved_vars()
        for ng in self._t_ampa_bufs:
            buf_len_ampa, buf_len_gaba = self._get_buf_lens_for_wslfp(self._wslfps[ng])
            self._t_ampa_bufs[ng] = deque(maxlen=buf_len_ampa)
            self._I_ampa_bufs[ng] = deque(maxlen=buf_len_ampa)
            self._t_gaba_bufs[ng] = deque(maxlen=buf_len_gaba)
            self._I_gaba[ng] = deque(maxlen=buf_len_gaba)



from __future__ import annotations

import warnings
from datetime import datetime
from typing import Callable

import matplotlib as mpl
import neo
import quantities as pq
from attrs import define, field, fields
from brian2 import NeuronGroup, Quantity, Unit, meter, mm, ms, np, um
from jaxtyping import Float, UInt
from matplotlib.artist import Artist
from mpl_toolkits.mplot3d import Axes3D

from cleo.base import Recorder
from cleo.coords import coords_from_ng
from cleo.imaging.sensors import Sensor
from cleo.utilities import (
    analog_signal,
    normalize_coords,
    rng,
    unit_safe_append,
    unit_safe_cat,
)


def target_neurons_in_plane(
    ng: NeuronGroup,
    scope_focus_depth: Quantity,
    scope_img_width: Quantity,
    scope_location: Quantity = (0, 0, 0) * mm,
    scope_direction: tuple = (0, 0, 1),
    soma_radius: Quantity = 10 * um,
    sensor_location: str = "cytoplasm",
) -> tuple[UInt[np.ndarray, "n_targets"], Float[np.ndarray, "n_targets"], Quantity]:
    """
    Returns a tuple of (i_targets, noise_focus_factor, coords_on_plane)

    Parameters
    ----------
    ng : NeuronGroup
        The neuron group to target.
    scope_focus_depth : Quantity
        The depth of the focal plane of the microscope.
    scope_img_width : Quantity
        The width of the image captured by the microscope.
    scope_location : Quantity, optional
        The location of the microscope, by default (0, 0, 0) * mm.
    scope_direction : tuple, optional
        The direction of the microscope, by default (0, 0, 1).
    soma_radius : Quantity, optional
        The radius of the soma of the neuron, by default 10 * um.
        Used to compute noise focus factor, since smaller ROIs will have
        a noisier distribution of fluorescence, averaged over fewer pixels.
    sensor_location : str, optional
        The location of the sensor, by default "cytoplasm".

    Returns
    -------
    Tuple[UInt[np.ndarray, "n_targets"], Float[np.ndarray, "n_targets"], Quantity]
        A tuple of (i_targets, noise_focus_factor, coords_on_plane)
    """
    assert sensor_location in ("cytoplasm", "membrane")

    ng_coords = coords_from_ng(ng)
    # Compute the normal vector and the center of the plane
    plane_normal = scope_direction
    plane_center = scope_location + plane_normal * scope_focus_depth
    # Compute the distance of each neuron from the plane
    perp_distances = np.abs(np.dot(ng_coords - plane_center, plane_normal))
    noise_focus_factor = np.ones(ng.N)
    # signal falloff with shrinking cross-section (or circumference)
    # ignore numpy warning
    with np.errstate(invalid="ignore"):
        r_soma_visible = np.sqrt(soma_radius**2 - perp_distances**2)
    r_soma_visible[np.isnan(r_soma_visible)] = 0
    if sensor_location == "cytoplasm":
        relative_num_pixels = (r_soma_visible / soma_radius) ** 2
    else:
        relative_num_pixels = r_soma_visible / soma_radius
    with np.errstate(divide="ignore"):
        noise_focus_factor /= np.sqrt(relative_num_pixels)

    # get only neurons in view
    coords_on_plane = ng_coords - plane_normal * perp_distances[:, np.newaxis]
    plane_distances = np.sqrt(np.sum((coords_on_plane - plane_center) ** 2, axis=1))

    i_targets = np.flatnonzero(
        np.logical_and(r_soma_visible > 0, plane_distances < scope_img_width / 2)
    )

    return i_targets, noise_focus_factor[i_targets], coords_on_plane[i_targets]


@define(eq=False)
class Scope(Recorder):
    """Two-photon microscope.

    Injection kwargs
    ----------------
    rho_rel_generator : Callable[[int], Float[np.ndarray, "n"], optional
        A function assigning expression levels. Takes n as an arg, outputs float array.
        ``lambda n: np.ones(n)`` by default.
    focus_depth : Quantity, optional
        The depth of the focal plane, by default that of the scope.
    soma_radius : Quantity, optional
        The radius of the soma of the neuron, by default that of the scope.
        Used to compute noise focus factor, since smaller ROIs will have
        a noisier distribution of fluorescence, averaged over fewer pixels.
    """

    sensor: Sensor = field()
    img_width: Quantity = field()
    """The width (diameter) of the (circular) image captured by the microscope.
    Specified in distance units."""
    focus_depth: Quantity = None
    """The depth of the focal plane, with distance units"""
    location: Quantity = [0, 0, 0] * mm
    """Location of the objective lens."""
    direction: Float[np.ndarray, "3"] = field(
        default=(0, 0, 1), converter=normalize_coords
    )
    """Direction in which the microscope is pointing.
    By default straight down (`+z` direction)"""
    soma_radius: Quantity = field(default=10 * um)
    """Assumed radius of neurons, used to compute noise focus factor.
    Smaller neurons have noisier signals."""
    snr_cutoff: float = field(default=1)
    """SNR below which neurons are discarded.
    Applied only when focus_depth is not None"""
    dFF: list[Float[np.ndarray, "{self.n}"]] = field(
        factory=list, init=False, repr=False
    )
    """ΔF/F from every call to :meth:`get_state`.
    Shape is (n_samples, n_ROIs). Stored if :attr:`~cleo.InterfaceDevice.save_history`"""
    t: Quantity = field(factory=lambda: ms * [], init=False, repr=False)
    """Times at which sensor traces are recorded, with Brian units, stored if
    :attr:`~cleo.InterfaceDevice.save_history`"""

    neuron_groups: list[NeuronGroup] = field(factory=list, repr=False, init=False)
    """neuron groups the scope has been injected into, in order of injection"""
    i_targets_per_injct: list[UInt[np.ndarray, "*"]] = field(
        factory=list, repr=False, init=False
    )
    """targets of neurons selected from each injection"""
    sigma_per_injct: list[Float[np.ndarray, "*"]] = field(
        factory=list, repr=False, init=False
    )
    """`sigma_noise` of neurons selected from each injection"""
    focus_coords_per_injct: list[Float[np.ndarray, "*"]] = field(
        factory=list, repr=False, init=False
    )
    """coordinates on the focal plane of neurons selected from each injection"""
    rho_rel_per_injct: list[Float[np.ndarray, "*"]] = field(
        factory=list, repr=False, init=False
    )
    """relative expression levels of neurons selected from each injection"""

    @property
    def n(self) -> int:
        """Number of imaged ROIs"""
        return sum(len(i_t) for i_t in self.i_targets_per_injct)

    @property
    def sigma_noise(self) -> Float[np.ndarray, "{self.n}"]:
        """noise std dev (in terms of ΔF/F) for all targets, in order injected."""
        if self.n == 0:
            return np.array([])
        return np.concatenate(self.sigma_per_injct)

    @property
    def focus_coords(self) -> Quantity:
        """coordinates on the focal plane of all targets, in order injected."""
        if self.n == 0:
            return [] * mm
        return unit_safe_cat(self.focus_coords_per_injct)

    @property
    def dFF_1AP(self) -> Float[np.ndarray, "{self.n}"]:
        """dFF_1AP for all targets, in order injected. Varies with expression levels."""
        if self.n == 0:
            return np.array([])
        rho_rel = np.concatenate(self.rho_rel_per_injct)
        assert rho_rel.shape == (self.n,)
        return rho_rel * self.sensor.dFF_1AP

    def _init_saved_vars(self):
        if self.save_history:
            self.t = fields(type(self)).t.default.factory()
            self.dFF = fields(type(self)).dFF.default.factory()

    def _update_saved_vars(self, t, dFF):
        if self.save_history:
            self.t = unit_safe_append(self.t, t)
            self.dFF.append(dFF)

    def __attrs_post_init__(self):
        self._init_saved_vars()

    def reset(self, **kwargs) -> None:
        self._init_saved_vars()

    def target_neurons_in_plane(
        self, ng, focus_depth: Quantity = None, soma_radius: Quantity = None
    ) -> tuple[
        UInt[np.ndarray, "n_targets"],
        Float[np.ndarray, "n_targets"],
        Float[np.ndarray, "n_targets"],
    ]:
        """calls :func:`target_neurons_in_plane` with scope parameter defaults.
        `focus_depth` and `soma_radius` can be overridden here.

        Parameters
        ----------
        ng : NeuronGroup
            The neuron group to target.
        focus_depth : Quantity
            The depth of the focal plane, by default that of the microscope.
        soma_radius : Quantity, optional
            The radius of the soma of the neuron, by default that of the microscope.
            Used to compute noise focus factor, since smaller ROIs will have
            a noisier distribution of fluorescence, averaged over fewer pixels.

        Returns
        -------
        Tuple[UInt[np.ndarray, "n_targets"], Float[np.ndarray, "n_targets"], Float[np.ndarray, "n_targets"]]
            A tuple of (i_targets, noise_focus_factor, coords_on_plane)
        """
        focus_depth = focus_depth or self.focus_depth
        soma_radius = soma_radius or self.soma_radius
        return target_neurons_in_plane(
            ng,
            focus_depth,
            self.img_width,
            self.location,
            self.direction,
            soma_radius,
            self.sensor.location,
        )

    def get_state(self) -> Float[np.ndarray, "{self.n}"]:
        """Returns a 1D array of ΔF/F values for all targets, in order injected.

        Returns
        -------
        Float[np.ndarray, "{self.n}"]
            Fluorescence values for all targets
        """
        signal = []
        signal_per_ng = self.sensor.get_state()
        # sensor has just one signal for neuron group, not storing
        # separately for each injection, so we'll recover that here
        n_prev_targets_for_ng = {}
        for ng, i_targets in zip(self.neuron_groups, self.i_targets_per_injct):
            if ng.name not in signal_per_ng:
                raise RuntimeError(
                    f"Sensor {self.sensor.name} has no signal for neuron group {ng.name}."
                    " Did you forget to call inject_sensor_for_targets() after scope injections??"
                )
            subset_start = n_prev_targets_for_ng.get(ng, 0)
            subset_for_injct = slice(subset_start, subset_start + len(i_targets))
            signal.append(signal_per_ng[ng.name][subset_for_injct])
            n_prev_targets_for_ng[ng] = subset_start + len(i_targets)
        signal = np.concatenate(signal)
        noise = rng.normal(0, self.sigma_noise, len(signal))
        assert self.n == len(signal) == len(noise) == len(self.sigma_noise)

        state = signal + noise
        t_now = self.sim.network.t
        self._update_saved_vars(t_now, state)
        return signal + noise

    def connect_to_neuron_group(self, neuron_group: NeuronGroup, **kwparams) -> None:
        focus_depth = kwparams.get("focus_depth", self.focus_depth)
        soma_radius = kwparams.get("soma_radius", self.soma_radius)
        rho_rel_generator = kwparams.get("rho_rel_generator", lambda n: np.ones(n))
        if focus_depth:
            if "i_targets" in kwparams or "sigma_noise" in kwparams:
                raise ValueError(
                    "i_targets and sigma_noise can not be specified while focus_depth",
                    " is not None",
                )
            (
                i_targets,
                noise_focus_factor,
                focus_coords,
            ) = self.target_neurons_in_plane(neuron_group, focus_depth, soma_radius)
            base_sigma = kwparams.get("base_sigma", self.sensor.sigma_noise)
            sigma_noise = noise_focus_factor * base_sigma
            rho_rel = rho_rel_generator(len(i_targets))
            if self.sensor.dFF_1AP is not None:
                snr = rho_rel * self.sensor.dFF_1AP / sigma_noise
                i_targets = i_targets[snr > self.snr_cutoff]
                sigma_noise = sigma_noise[snr > self.snr_cutoff]
                focus_coords = focus_coords[snr > self.snr_cutoff]
                rho_rel = rho_rel[snr > self.snr_cutoff]
            else:
                warnings.warn(
                    f"SNR cutoff not used, since {self.sensor.name} does not have dFF_1AP defined."
                )
        else:
            i_targets = kwparams.pop("i_targets", neuron_group.i_)
            sigma_noise = kwparams.get("sigma_noise", self.sensor.sigma_noise)
            rho_rel = rho_rel_generator(len(i_targets))
            _, sigma_noise, rho_rel = np.broadcast_arrays(
                i_targets, sigma_noise, rho_rel
            )
            focus_coords = coords_from_ng(neuron_group)[i_targets]
        assert len(i_targets) == len(sigma_noise) == len(focus_coords) == len(rho_rel)

        self.neuron_groups.append(neuron_group)
        self.i_targets_per_injct.append(i_targets)
        self.sigma_per_injct.append(sigma_noise)
        self.focus_coords_per_injct.append(focus_coords)
        self.rho_rel_per_injct.append(rho_rel)

    def i_targets_for_neuron_group(self, neuron_group):
        """can handle multiple injections into same ng"""
        i_targets_for_ng = []
        for ng, i_targets in zip(self.neuron_groups, self.i_targets_per_injct):
            if ng is neuron_group:
                i_targets_for_ng.extend(i_targets)
        return i_targets_for_ng

    def inject_sensor_for_targets(self, **kwparams) -> None:
        for ng in set(self.neuron_groups):
            i_targets_for_ng = []
            rho_rel_for_ng = []
            for ng_, i_targets, rho_rel in zip(
                self.neuron_groups, self.i_targets_per_injct, self.rho_rel_per_injct
            ):
                if ng_ is not ng:
                    continue

                i_targets_for_ng.extend(i_targets)
                rho_rel_for_ng.extend(rho_rel)

            self.sim.inject(
                self.sensor,
                ng,
                i_targets=i_targets_for_ng,
                rho_rel=rho_rel_for_ng,
                **kwparams,
            )

    def add_self_to_plot(
        self, ax: Axes3D, axis_scale_unit: Unit, **kwargs
    ) -> list[Artist]:
        color = kwargs.pop("color", "xkcd:fluorescent green")
        snr = np.concatenate(self.sigma_per_injct)
        coords = (
            np.concatenate(
                [
                    coords_from_ng(ng)[i_targets] / meter
                    for ng, i_targets in zip(
                        self.neuron_groups, self.i_targets_per_injct
                    )
                ]
            )
            * meter
        )
        assert coords.shape == (len(snr), 3)
        assert len(snr) == len(coords)
        scope_marker = ax.quiver(
            self.location[0] / axis_scale_unit,
            self.location[1] / axis_scale_unit,
            self.location[2] / axis_scale_unit,
            self.direction[0],
            self.direction[1],
            self.direction[2],
            color=color,
            lw=5,
            label=self.name,
            pivot="tail",
            length=self.focus_depth / axis_scale_unit,
            normalize=True,
        )

        # Define the center, normal vector, and radius
        normal = self.direction
        center = self.location + normal * self.focus_depth
        radius = self.img_width / 2

        # Generate the points for the circle
        theta = np.linspace(0, 2 * np.pi, 100)
        r = np.sqrt(np.linspace(0, 1, 5))
        # angle away from z axis
        phi = np.arccos(normal[2])
        # angle relative to x axis
        theta0 = np.arctan2(normal[1], normal[0])
        x = center[0] + radius * np.cos(phi) * np.outer(np.cos(theta), r)
        y = center[1] + radius * np.cos(phi) * np.outer(np.sin(theta), r)
        z = center[2] + radius * np.sin(phi) * np.outer(-np.cos(theta - theta0), r)

        plane = ax.plot_surface(
            x / axis_scale_unit,
            y / axis_scale_unit,
            z / axis_scale_unit,
            color=color,
            alpha=0.2,
        )

        target_markers = ax.scatter(
            coords[:, 0] / axis_scale_unit,
            coords[:, 1] / axis_scale_unit,
            coords[:, 2] / axis_scale_unit,
            marker="^",
            c=color,
            label=f"{self.sensor.name} ROIs",
            **kwargs,
        )
        color_rgba = target_markers.get_facecolor()
        color_rgba[:, :3] = 0.3 * color_rgba[:, :3]
        target_markers.set(color=color_rgba)

        handles = ax.get_legend().legend_handles
        handles.append(target_markers)
        patch = mpl.patches.Patch(color=color, label=self.name)
        handles.append(patch)
        ax.legend(handles=handles)

        return [scope_marker, target_markers, plane]

    def to_neo(self) -> neo.AnalogSignal:
        # inherit docstring
        try:
            signal = analog_signal(self.t, self.dFF, "")
        except AttributeError:
            return
        signal.name = self.name
        signal.description = (
            "Fluorescence trace for each ROI. Exported from Cleo Scope device"
        )
        signal.annotate(
            export_datetime=datetime.now(),
            sensor=self.sensor.name,
            scope_direction=self.direction,
            scope_location=self.location / um * pq.um,
            focus_depth=self.focus_depth / um * pq.um,
            img_width=self.img_width / um * pq.um,
        )
        if self.n > 0:
            focus_coords = self.focus_coords / um * pq.um
            signal.array_annotate(
                dFF_1AP=self.dFF_1AP,
                x=focus_coords[..., 0],
                y=focus_coords[..., 1],
                z=focus_coords[..., 2],
                i_roi=np.arange(self.n),
            )
        return signal
        # want ROI coordinates and size in image



"""Contains Scope and sensors for two-photon microscopy"""
from cleo.imaging.scope import Scope, target_neurons_in_plane
from cleo.imaging.sensors import (
    GECI,
    CalBindingActivationModel,
    CalciumModel,
    DoubExpCalBindingActivation,
    DynamicCalcium,
    ExcitationModel,
    LightExcitation,
    NullBindingActivation,
    NullExcitation,
    PreexistingCalcium,
    Sensor,
    gcamp3,
    gcamp6_rs06,
    gcamp6_rs09,
    gcamp6f,
    gcamp6s,
    geci,
    jgcamp7b,
    jgcamp7c,
    jgcamp7f,
    jgcamp7s,
    ogb_1,
)
